from __future__ import absolute_import

import io
import string
import uuid

import pytest

import pike.ntstatus
import pike.smb2


@pytest.fixture
def filename():
    return "test_io_{}.bin".format(uuid.uuid4())


@pytest.fixture
def test_buffer():
    def buf(size):
        base = string.printable
        copies = size // len(base) + 1
        return (base * copies).encode("utf-8")[:size]

    return buf


def test_open_write_read_seek(pike_TreeConnect, filename, test_buffer):
    with pike_TreeConnect() as tc:
        with tc.chan.create(
            tc.tree,
            filename,
            access=pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE | pike.smb2.DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result() as fh:
            # write a buffer greater than the transact size
            max_write = tc.conn.negotiate_response.max_write_size
            buf = test_buffer(size=max_write * 3)
            fh.write(buf)

            # make sure flush works
            flush_resp = fh.flush()
            assert flush_resp
            assert flush_resp.status == pike.ntstatus.STATUS_SUCCESS

            # the EOF marker should have been updated
            assert fh.end_of_file == len(buf)

            # now we're at the end of file, so read should return empty
            assert fh.read() == b""

            # subsequent reads should also return empty
            assert fh.read() == b""

            # seek backwards to read 8 bytes
            pos = fh.seek(-8, io.SEEK_CUR)
            assert pos == fh.tell()

            tail_buf_8 = fh.read()
            assert len(tail_buf_8) == 8
            assert tail_buf_8 == buf[-8:]
            assert fh.tell() == pos + 8

            # seek to beginning of file
            pos = fh.seek(0, io.SEEK_SET)
            assert pos == 0

            # read to the end
            read_buf = fh.read()
            assert read_buf == buf
            assert fh.tell() == fh.end_of_file

            # seek -256 from the end
            pos = fh.seek(-256, io.SEEK_END)
            assert pos == fh.tell()
            assert pos == fh.end_of_file - 256
            tail_buf_256_16 = fh.read(16)
            assert len(tail_buf_256_16) == 16
            assert tail_buf_256_16 == buf[-256:-240]
            assert fh.tell() == pos + 16

            # seek backwards 512
            pos = fh.seek(-512, io.SEEK_CUR)
            assert pos == fh.tell()
            assert pos == fh.end_of_file - 256 + 16 - 512
            tail_buf_752_256 = fh.read(256)
            assert len(tail_buf_752_256) == 256
            assert tail_buf_752_256 == buf[-752:-496]
            assert fh.tell() == pos + 256


def test_open_not_readable_writable(pike_TreeConnect, filename):
    with pike_TreeConnect() as tc:
        with tc.chan.create(
            tc.tree,
            filename,
            access=pike.smb2.GENERIC_READ | pike.smb2.DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result() as rfh:
            assert not rfh.writable()
            # can't flush non-writable streams
            assert rfh.flush() is None

        with tc.chan.create(
            tc.tree,
            filename,
            access=pike.smb2.GENERIC_WRITE | pike.smb2.DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result() as wfh:
            assert not wfh.readable()


def test_update_eof(pike_TreeConnect, filename, test_buffer):
    with pike_TreeConnect() as tc:
        with tc.chan.create(
            tc.tree,
            filename,
            access=pike.smb2.GENERIC_READ | pike.smb2.DELETE,
            share=pike.smb2.FILE_SHARE_ALL,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result() as rfh:
            with tc.chan.create(
                tc.tree,
                filename,
                access=pike.smb2.GENERIC_WRITE,
                share=pike.smb2.FILE_SHARE_ALL,
            ).result() as wfh:
                assert rfh.end_of_file == wfh.end_of_file
                buf = test_buffer(size=256)
                wfh.write(buf)
                # the two handles eof will get out of sync at this point
                assert wfh.end_of_file == len(buf)
                assert rfh.end_of_file != wfh.end_of_file

                # reading to an offset should update the eof, but they're still not equal
                assert rfh.read(128) == buf[:128]
                assert rfh.end_of_file == 128
                assert rfh.end_of_file != wfh.end_of_file

                # reading to end should update the eof, now they're in sync again
                assert rfh.read() == buf[128:]
                assert rfh.end_of_file == wfh.end_of_file


def test_open_write_multiple(pike_TreeConnect, filename, test_buffer):
    with pike_TreeConnect() as tc:
        with tc.chan.create(
            tc.tree,
            filename,
            access=pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE | pike.smb2.DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result() as fh:
            # write a buffer greater than the transact size
            max_write = tc.conn.negotiate_response.max_write_size
            buf = test_buffer(256)
            n_ops = 5
            for i in range(n_ops):
                fh.write(buf)
            fh.seek(0)
            for i in range(n_ops):
                assert fh.read(256) == buf



"""
Path-like interface for working with a Tree object
"""
from __future__ import absolute_import

import datetime
import io
import os
import sys
import warnings

try:
    from pathlib import PureWindowsPath, _WindowsFlavour
except ImportError:
    from pathlib2 import PureWindowsPath, _WindowsFlavour

from . import model
from . import ntstatus
from . import nttime
from . import smb2


MAX_SYMLINK_RECURSE = 10

NOT_FOUND_STATUSES = (
    ntstatus.STATUS_OBJECT_NAME_NOT_FOUND,
    ntstatus.STATUS_OBJECT_PATH_NOT_FOUND,
)


class _PikeFlavour(_WindowsFlavour):
    """
    Implement pike-specific overrides of WindowsFlavour semantics.

    Handle the fact that ``_drv`` may be a :py:class:`pike.model.Tree` instance.
    """

    def _fspath(self, p):
        if sys.version_info < (3,):
            return p.__fspath__()
        return os.fspath(p)

    def casefold(self, s):
        if isinstance(s, model.Tree):
            s = self._fspath(s)
        return s.lower()

    def casefold_parts(self, parts):
        return [self.casefold(rp) for rp in parts]


_pike_flavour = _PikeFlavour()


class PikePath(PureWindowsPath):
    """
    ``Path`` subclass for :py:class:`pike.model.Tree` instances.

    ``PikePath`` should be either directly instantiated by passing a
    :py:class:`~pike.model.Tree` or by performing typical ``Path`` extension
    (truediv) on a :py:class:`~pike.model.Tree` or
    :py:class:`pike.TreeConnect` instance.

    .. code-block:: python
        :caption: example.py

        import pike.path
        import pike.test

        with pike.TreeConnect() as tc:
            pth = pike.path.PikePath(tc.tree)
            assert (tc / "subdir") == (pth / "subdir")
            assert (tc.tree / "subdir") == (pth / "subdir")

    A ``PikePath`` is bound to a specific connection, session, and tree. If the
    underlying transport is no longer valid, then the path operations will not
    work.

    An existing path, valid or invalid, and be "rehomed" to a different
    ``PikePath`` by calling :py:func:`~pike.path.PikePath.join_from_root`,
    returning a new instance with the path part replaced by that of the passed
    argument.

    .. code-block:: python
        :caption: example.py

        import pike
        import pike.path

        with pike.TreeConnect(), pike.TreeConnect() as tc1, tc2:
            tc1_subdir = tc1 / "subdir"
            tc2_subdir = pike.path.PikePath(tc2.tree).join_from_root(tc1_subdir)
            assert tc1_subdir != tc2_subdir

    Unlike ``WindowsPath``, ``PikePath`` will prefer to raise
    :py:class:`pike.model.ResponseError` exceptions instead of ``OSError``.
    """

    _flavour = _pike_flavour

    @classmethod
    def _parse_args(cls, args):
        """
        Override _parse_args to ensure a Tree is returned as drv
        """
        if sys.version_info < (3,):
            # On python 2, explicitly convert future.types.newstr to native unicode type.
            from future.types import newstr

            # ensure that `newstr` instances are passed as real `unicode`
            # to avoid TypeError: can't intern subclass of string
            args = tuple(unicode(pc) if isinstance(pc, newstr) else pc for pc in args)
        drv, root, parts = super(PikePath, cls)._parse_args(args)
        if args:
            if isinstance(args[0], model.Tree):
                drv = args[0]
            if isinstance(args[0], cls):
                drv = args[0]._drv
        return drv, root, parts

    @property
    def _channel(self):
        """
        :rtype: pike.model.Channel
        :return: Channel bound to the associated tree
        """
        return self._drv.session.first_channel()

    @property
    def drive(self):
        """
        :rtype: pike.model.Tree
        :return: Tree forming the root of the share / path
        """
        return super(PikePath, self).drive

    _tree = drive

    @property
    def _path(self):
        """
        :rtype: str
        :return: SMB path relative to the root
        """
        if self._drv or self._root:
            return self._flavour.join(self._parts[1:])
        return str(self)

    def join_from_root(self, path):
        """
        Rehome an existing absolute path for use by this PikePath's Tree.

        :rtype: PikePath
        :return: a new :py:class:`~pike.path.PikePath` by joining ``path`` to
            the root of this instance.
        """
        if isinstance(path, type(self)) and (path.is_absolute()):
            return self.joinpath(*path._parts[1:])
        else:
            path = type(self)("\\", path)
        return self / path

    @classmethod
    def cwd(cls):
        raise NotImplementedError("No concept of cwd for {!r}".format(cls))

    @classmethod
    def home(cls):
        raise NotImplementedError("No concept of home for {!r}".format(cls))

    @staticmethod
    def _follow_link(link, reparse_data):
        """Return the next target for link by following reparse_data."""
        if reparse_data.unparsed_path_length > 0:
            prefix = link.join_from_root(
                link._path[: -reparse_data.unparsed_path_length // 2]
            ).parent
            unparsed_portion = link._path[-reparse_data.unparsed_path_length // 2 :]
        else:
            prefix = link.parent
            unparsed_portion = ""
        return prefix / (reparse_data.substitute_name + unparsed_portion)

    def _create(self, *args, **kwargs):
        """
        Call Channel.create with this Tree and path following symlinks.

        :param _follow: if False, raise symlink errors instead of following

        Special kwarg, __RC_DEPTH, is used to track the recursion depth and
        will raise OSError if called more than MAX_SYMLINK_RECURSE times.

        Remaining args and kwargs are passed to create directly.
        """
        depth = kwargs.pop("__RC_DEPTH", 0)
        if depth > MAX_SYMLINK_RECURSE:
            raise OSError("Maximum level of symbolic links exceeded")
        resolved_path = self
        if ".." in self.parts or "." in self.parts:
            resolved_path = self.resolve_relative()
        follow = kwargs.pop("_follow", True)
        try:
            return self._channel.create(
                self._tree, resolved_path._path, *args, **kwargs
            ).result()
        except model.ResponseError as re:
            if re.response.status != ntstatus.STATUS_STOPPED_ON_SYMLINK or not follow:
                raise
            kwargs["__RC_DEPTH"] = depth + 1
            return self._follow_link(
                link=resolved_path,
                reparse_data=re.response[0][0].error_data,
            )._create(*args, **kwargs)

    def stat(
        self,
        file_information_class=smb2.FILE_BASIC_INFORMATION,
        info_type=smb2.SMB2_0_INFO_FILE,
        options=0,
    ):
        """
        Perform SMB2 QUERY_INFO request for the given ``file_information_class``.

        :type file_information_class: smb2.FileInformationClass
        :param file_information_class: the file information class constant
            describing the type of information to retrieve
        :type info_type: smb2.InfoType
        :param info_type: the SMB2 query info type
        :type options: smb2.CreateOptions
        :param options: options passed to CREATE when opening the file
        :return: single file info object corresponding to the
            file_information_class requested
        """
        with self._create(
            access=smb2.FILE_READ_ATTRIBUTES,
            disposition=smb2.FILE_OPEN,
            options=options,
        ) as handle:
            return self._channel.query_file_info(
                handle, file_information_class, info_type, first_result_only=True
            )

    def lstat(
        self,
        file_information_class=smb2.FILE_BASIC_INFORMATION,
        info_type=smb2.SMB2_0_INFO_FILE,
        options=0,
    ):
        """
        Perform SMB2 QUERY_INFO request for the given ``file_information_class``.

        If this path represents a symlink, open the link itself, rather than the
        target. (Intermediate links will not be followed first.)

        :type file_information_class: smb2.FileInformationClass
        :param file_information_class: the file information class constant
            describing the type of information to retrieve
        :type info_type: smb2.InfoType
        :param info_type: the SMB2 query info type
        :type options: smb2.CreateOptions
        :param options: options passed to CREATE when opening the file
        :return: single file info object corresponding to the
            file_information_class requested
        """
        options = options | smb2.FILE_OPEN_REPARSE_POINT
        return self.stat(
            file_information_class=file_information_class,
            info_type=info_type,
            options=options,
        )

    def chmod(self, mode):
        raise NotImplementedError("ACL modification is not supported")

    lchmod = chmod

    def exists(self, options=0):
        """
        :type options: smb2.CreateOptions
        :param options: options passed to CREATE when opening the file
        :return: True if this path exists
        """
        try:
            with self._create(
                access=smb2.FILE_READ_ATTRIBUTES,
                disposition=smb2.FILE_OPEN,
                options=options,
            ) as handle:
                return True
        except model.ResponseError as re:
            if re.response.status not in NOT_FOUND_STATUSES + (
                ntstatus.STATUS_FILE_IS_A_DIRECTORY,
                ntstatus.STATUS_NOT_A_DIRECTORY,
            ):
                raise
        return False

    def expanduser(self):
        raise NotImplementedError("expanduser() is not supported")

    def group(self):
        raise NotImplementedError("group() is not supported")

    def is_dir(self):
        """
        :return: True if this path is a directory
        """
        return self.exists(options=smb2.FILE_DIRECTORY_FILE)

    def is_file(self):
        """
        :return: True if this path is a normal file
        """
        return self.exists(options=smb2.FILE_NON_DIRECTORY_FILE)

    def is_symlink(self):
        """
        :return: True if this path is a reparse point
        """
        try:
            with self._create(
                access=smb2.FILE_READ_ATTRIBUTES,
                disposition=smb2.FILE_OPEN,
                _follow=False,
            ) as handle:
                pass
        except model.ResponseError as re:
            if re.response.status == ntstatus.STATUS_STOPPED_ON_SYMLINK:
                return True
            if re.response.status not in NOT_FOUND_STATUSES:
                raise
        return False

    def is_mount(self):
        """
        Note: all PikePath instances are considered to be mounts

        :return: True if this path is a mount point
        """
        return True  # all paths are on SMB mount

    def is_socket(self):
        """
        Note: no PikePath instances are considered to be sockets

        :return: True if this path is a socket
        """
        return False

    def is_fifo(self):
        """
        Note: no PikePath instances are considered to be fifo

        :return: True if this path is a pipe
        """
        return False  # might be supported some day

    def is_block_device(self):
        """
        Note: no PikePath instances are considered to be block devices

        :return: True if this path is a block device
        """
        return False

    def is_char_device(self):
        """
        Note: no PikePath instances are considered to be char devices

        :return: True if this path is a char device
        """
        return False

    def glob(self, pattern):
        """
        Iterate over this subtree and yield all existing files matching pattern.

        Files of any kind, including directories, that match the given relative
        pattern will be yielded with the exception of special entries "." and ".."

        The globbing is handled on the server side, so all ``pattern`` values
        must be valid according to [MS-CIFS] 2.2.1.1.3 Wildcards:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cifs/dc92d939-ec45-40c8-96e5-4c4091e4ab43

        The recursive glob "**" is not supported.

        :param pattern: file search pattern
        :rtype: Iterator[:py:class:`~pike.path.PikePath`]
        """
        if "**" in pattern:
            raise ValueError(
                "recursive glob, '**', is not supported by {!r}".format(type(self))
            )
        with self._create(
            access=smb2.GENERIC_READ,
            disposition=smb2.FILE_OPEN,
            options=smb2.FILE_DIRECTORY_FILE,
        ) as handle:
            for item in handle.enum_directory(
                file_information_class=smb2.FILE_NAMES_INFORMATION, file_name=pattern
            ):
                if item.file_name in (".", ".."):
                    continue
                yield self / item.file_name

    def iterdir(self):
        """
        Iterate over this subtree and yield all existing files.

        The special entries "." and ".." will not be yielded.

        :rtype: Iterator[:py:class:`~pike.path.PikePath`]
        """

        for item in self.glob("*"):
            yield item

    def link_to(self, target):
        raise NotImplementedError("hardlinks are not supported over SMB")

    def mkdir(self, mode=None, parents=False, exist_ok=False):
        """
        Create a new directory at this given path.

        :param mode: The mode of the created file. NOT SUPPORTED
        :param parents: if True, create all parent directories if they do not exist
        :param exist_ok: if True, don't raise if the directory exists
        """
        if mode is not None:
            warnings.warn("`mode` argument is not handled at this time", stacklevel=2)
        try:
            with self._create(
                access=smb2.GENERIC_WRITE,
                disposition=smb2.FILE_CREATE,
                options=smb2.FILE_DIRECTORY_FILE,
            ) as handle:
                return
        except model.ResponseError as re:
            if re.response.status == ntstatus.STATUS_OBJECT_NAME_COLLISION and exist_ok:
                return
            if not parents or self.parent == self:
                raise
            self.parent.mkdir(parents=True, exist_ok=True)
            self.mkdir(mode, parents=False, exist_ok=exist_ok)

    def open(self, mode="r", buffering=-1, encoding=None, errors=None, newline=None):
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.

        Where possible, the semantics are designed to match a local open file.

        Lease/oplock based buffering is not supported at this time. To avoid
        reading stale data, or locally buffering writes, pass ``buffering=0``.
        """
        buffer_class = io.BufferedReader
        access = 0
        disposition = 0
        mode = mode.lower()
        if ("r" in mode, "w" in mode, "x" in mode, "a" in mode).count(True) > 1:
            raise ValueError("must have exactly one of create/read/write/append mode")

        # File Access bits
        if "r" in mode or "+" in mode:
            access |= smb2.GENERIC_READ
        if "a" in mode or "x" in mode or "w" in mode or "+" in mode:
            access |= smb2.GENERIC_WRITE
            buffer_class = io.BufferedWriter
        if "+" in mode:
            buffer_class = io.BufferedRandom

        # Create disposition bits
        if "r" in mode:
            disposition = smb2.FILE_OPEN
        elif "x" in mode:
            disposition = smb2.FILE_CREATE
        elif "a" in mode:
            disposition = smb2.FILE_OPEN_IF
        elif "w" in mode:
            disposition = smb2.FILE_SUPERSEDE

        handle = self._create(
            access=access,
            disposition=disposition,
            options=smb2.FILE_NON_DIRECTORY_FILE,
        )
        if "a" in mode:
            handle.seek(0, io.SEEK_END)
        if "b" in mode and buffering == 0:
            return handle
        if buffering == 0:
            warnings.warn("Buffering cannot be disabled for text-mode streams")
        buffer_size = buffering if buffering > 0 else smb2.BYTES_PER_CREDIT
        buffered_io = buffer_class(handle, buffer_size=buffer_size)
        if "b" in mode:
            return buffered_io
        return io.TextIOWrapper(
            buffered_io, encoding=encoding, errors=errors, newline=newline
        )

    def read_bytes(self):
        """
        Return the binary contents of the pointed-to file as a bytes object.

        :rtype: bytes
        """
        with self.open("rb") as f:
            return f.read()

    def read_text(self, encoding=None, errors=None):
        """
        Return the decoded contents of the pointed-to file as a string.

        :rtype: str
        """
        with self.open("r") as f:
            return f.read()

    def readlink(self):
        """
        Return the :py:class:`~pike.path.PikePath` to which the symbolic link points.

        :rtype: PikePath
        """
        with self._create(
            access=smb2.FILE_READ_ATTRIBUTES,
            disposition=smb2.FILE_OPEN,
            options=smb2.FILE_NON_DIRECTORY_FILE | smb2.FILE_OPEN_REPARSE_POINT,
        ) as handle:
            # XXX: not making additional tree connects here, so absolute links
            # will only work across the same share
            return self._follow_link(
                link=self,
                reparse_data=handle.get_symlink()[0][0][0],
            )

    def rename(self, target, replace=True):
        """
        Rename the object at this path to ``target``.

        :param target: new name
        :param replace: if True, pass ``replace_if_exists`` in ``FileRenameInformation``
        :rtype: PikePath
        :return: path pointing to target
        """
        if not isinstance(target, type(self)):
            target = self.join_from_root(target)
        with self._create(
            access=smb2.DELETE,
            disposition=smb2.FILE_OPEN,
        ) as handle:
            with handle.set_file_info(smb2.FileRenameInformation) as file_info:
                file_info.replace_if_exists = replace
                file_info.file_name = target._path
            return target

    def replace(self, target):
        """
        Same as rename, but with replace=True
        """
        return self.rename(target, replace=True)

    def resolve_relative(self):
        """Remove "." and ".." components from the path."""
        path_components = []
        for pc in self.parts[1:]:
            if pc == ".":
                continue
            if pc == "..":
                path_components.pop()
                continue
            path_components.append(pc)
        return self.join_from_root(type(self)(*path_components))

    def resolve(self, strict=True):
        """
        Follow all symbolic links in the path and return the
        :py:class:`~pike.path.PikePath` at the end of the chain.

        Similar to ``readlink``, however this calls relies on handling the
        SymbolicLinkErrorResponse during create, rather than explicitly
        reading the link target via IOCTL_GET_REPARSE_POINT.
        """
        with self._create(
            access=smb2.FILE_READ_ATTRIBUTES,
            disposition=smb2.FILE_OPEN,
        ) as handle:
            return self.join_from_root(handle.create_request.name)

    def rglob(self, pattern):
        raise NotImplementedError("rglob is not supported by {!r}".format(type(self)))

    def rmdir(self, missing_ok=False):
        """
        Remove the directory at this path.

        The directory must be empty.

        :param missing_ok: if True, don't raise if the directory doesn't exist
        """
        self.unlink(missing_ok=missing_ok, options=smb2.FILE_DIRECTORY_FILE)

    def samefile(self, otherpath):
        """
        Determine if this path points to the same file as another path.

        This uses ``FILE_INTERNAL_INFORMATION`` to perform the check. If the
        server doesn't support this info class, NotImplementedError is raised.

        If either path doesn't exist, return False

        :type otherpath: PikePath
        :param otherpath: the other path to compare to
        :return: True if otherpath points to the same file
        """
        try:
            my_id = self.stat(file_information_class=smb2.FILE_INTERNAL_INFORMATION)
        except model.ResponseError as re:
            if re.response.status in NOT_FOUND_STATUSES:
                return False
            raise
        if my_id == 0:
            raise NotImplementedError(
                "remote filesystem does not return unique file index"
            )
        if not isinstance(otherpath, type(self)):
            otherpath = self.join_from_root(otherpath)
        try:
            other_id = otherpath.stat(
                file_information_class=smb2.FILE_INTERNAL_INFORMATION
            )
        except model.ResponseError as re:
            if re.response.status in NOT_FOUND_STATUSES:
                return False
            raise
        return my_id.index_number == other_id.index_number

    def relative_root(self):
        """
        Return a PureWindowsPath which joined to this path would be the root.
        """
        path_components = []
        for pc in self.parts[1:]:
            if pc == "..":
                path_components.pop()
                continue
            path_components.append("..")
        return PureWindowsPath(*path_components)

    def symlink_to(self, target, target_is_directory=False):
        """
        Make this path a symlink pointing to the given path.

        Note the order of arguments (self, target) is the reverse of os.symlink's.

        Does not work for absolute links (across servers).

        :type target: PikePath
        :param target: The link's target
        :param target_is_directory: must be True if the target is a directory
        """
        if not isinstance(target, type(self)):
            target = self.join_from_root(target)
        try:
            substitute_path = target.relative_to(self.parent)._path
        except ValueError:
            # no common subpath, so relpath to the share root
            substitute_path = str(self.parent.relative_root() / target._path)

        options = smb2.FILE_DIRECTORY_FILE if target_is_directory else 0
        with self._create(
            access=smb2.GENERIC_WRITE,
            attributes=smb2.FILE_ATTRIBUTE_REPARSE_POINT,
            disposition=smb2.FILE_SUPERSEDE,
            options=options | smb2.FILE_OPEN_REPARSE_POINT,
        ) as handle:
            handle.set_symlink(substitute_path, flags=smb2.SYMLINK_FLAG_RELATIVE)

    def touch(self, mode=None, exist_ok=True):
        """
        Create a new file or update modification time at this given path.

        :param mode: The mode of the created file. NOT SUPPORTED
        :param exist_ok: if True, don't raise if the file exists
        """
        if mode is not None:
            warnings.warn("`mode` argument is not handled at this time", stacklevel=2)
        disposition = smb2.FILE_OPEN_IF if exist_ok else smb2.FILE_CREATE
        with self._create(
            access=smb2.GENERIC_WRITE,
            disposition=disposition,
            options=smb2.FILE_NON_DIRECTORY_FILE,
        ) as handle:
            with handle.set_file_info(smb2.FileBasicInformation) as file_info:
                file_info.change_time = nttime.NtTime(datetime.datetime.now())

    def unlink(self, missing_ok=False, options=smb2.FILE_NON_DIRECTORY_FILE):
        """
        Remove the file at this path.

        :param missing_ok: if True, don't raise if the file doesn't exist
        :param options: options passed to CREATE when opening the file
        """
        try:
            with self._create(
                access=smb2.DELETE,
                disposition=smb2.FILE_OPEN,
                options=options | smb2.FILE_DELETE_ON_CLOSE,
                _follow=False,
            ) as handle:
                pass
        except model.ResponseError as re:
            if re.response.status in NOT_FOUND_STATUSES and missing_ok:
                return
            elif re.response.status == ntstatus.STATUS_STOPPED_ON_SYMLINK:
                return self.unlink(
                    missing_ok=missing_ok,
                    options=options | smb2.FILE_OPEN_REPARSE_POINT,
                )
            raise

    def write_bytes(self, data):
        """
        Open the file pointed to in bytes mode, write data to it, and close the file.

        :type data: bytes
        :param data: bytes to write to the file
        :rtype: int
        :return: number of bytes written
        """
        with self.open("wb") as f:
            return f.write(data)

    def write_text(self, data, encoding=None, errors=None):
        """
        Open the file pointed to in text mode, write data to it, and close the file.

        :type data: str
        :param data: unicode string to write to the file
        :rtype: int
        :return: number of bytes written
        """
        with self.open("w") as f:
            return f.write(data)



#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#

import pytest

from pike.model import Client
from pike.smb2 import Dialect


@pytest.mark.parametrize(
    "min_dialect,max_dialect",
    (
        pytest.param(None, None, id="None/None"),
        pytest.param(Dialect.DIALECT_SMB2_1, None, id="lower bound"),
        pytest.param(None, Dialect.DIALECT_SMB2_1, id="upper bound"),
        pytest.param(Dialect.DIALECT_SMB3_1_1, Dialect.DIALECT_SMB2_002, id="inverted"),
        pytest.param(Dialect.DIALECT_SMB3_0, float("inf"), id="upper bound inf"),
        pytest.param(Dialect.DIALECT_SMB3_0, 0xFFFFFFFF, id="upper bound int"),
        pytest.param(float("-inf"), Dialect.DIALECT_SMB2_002, id="lower bound -inf"),
        pytest.param(0, Dialect.DIALECT_SMB3_0_2, id="lower bound zero"),
    ),
)
def test_restrict_dialect(min_dialect, max_dialect):
    c = Client()
    c.restrict_dialects(min_dialect, max_dialect)
    print(
        "restrict_dialects(min_dialect={}, max_dialect={}) => {}".format(
            min_dialect, max_dialect, c.dialects
        )
    )
    for d in c.dialects:
        if min_dialect is None:
            min_dialect = min(Dialect.values())
        if max_dialect is None:
            max_dialect = max(Dialect.values())
        assert d >= min_dialect
        assert d <= max_dialect



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        netbios.py
#
# Abstract:
#
#        NETBios frame support
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

from __future__ import absolute_import

from . import core
from . import crypto
from . import smb2


class Netbios(core.Frame):
    LOG_CHILDREN_COUNT = False
    LOG_CHILDREN_EXPAND = True

    def __init__(self, context=None):
        core.Frame.__init__(self, None, context)
        self.len = None
        self.conn = context
        self.transform = None
        self._smb2_frames = []

    def _log_str(self):
        components = []
        if self.transform:
            return self.transform._log_str()
        if not self.children:
            components.append(type(self).__name__)
        else:
            components.append(", ".join(self._log_str_children()))
        return " ".join(components)

    def _children(self):
        return self._smb2_frames

    def _encode(self, cur):
        # Frame length (0 for now)
        len_hole = cur.hole.encode_uint32be(0)
        base = cur.copy()
        if self.transform is not None:
            # performing encryption
            self.transform.encode(cur)
        else:
            for child in self.children:
                child.encode(cur)

        if self.len is None:
            self.len = cur - base
        len_hole(self.len)

    def _decode(self, cur):
        self.len = cur.decode_uint32be()
        end = cur + self.len

        with cur.bounded(cur, end):
            while cur < end:
                signature = cur.copy().decode_bytes(4)
                if signature.tobytes() == b"\xfdSMB":
                    crypto.TransformHeader(self).decode(cur)
                else:
                    smb2.Smb2(self).decode(cur)

    def append(self, smb2_frame):
        self._smb2_frames.append(smb2_frame)

    def adopt(self, child, related=True):
        """
        become the parent of child

        :param related: if True and child is an Smb2 frame, set the flag
                        SMB2_FLAGS_RELATED_OPERATIONS
        """
        self.append(child)
        child.parent = self
        if related and isinstance(child, smb2.Smb2):
            child.flags |= smb2.SMB2_FLAGS_RELATED_OPERATIONS



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        tree.py
#
# Abstract:
#
#        Tree tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.test


class TreeTest(pike.test.PikeTest):
    # Connect a tree, then disconnect it
    def test_tree(self):
        chan, tree = self.tree_connect()

        chan.tree_disconnect(tree)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        core.py
#
# Abstract:
#
#        Core API
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#          Masen Furer (masen.furer@dell.com)
#

"""
Core Pike infrastructure
"""

from builtins import hex
from builtins import str
from builtins import object
import array
from binascii import hexlify
import struct

from future.utils import with_metaclass


class BufferOverrun(Exception):
    """Buffer overrun exception"""

    def __init__(self, cursor, request, boundary):
        self.cursor = cursor
        self.request = request
        self.boundary = boundary

    @property
    def message(self):
        return "{} is {} bound {} in {!r}\nPacket buffer: {!r}".format(
            self.request,
            "below lower" if self.request < self.boundary else "above upper",
            self.boundary,
            self.cursor,
            hexlify(self.cursor.array),
        )

    def __str__(self):
        return self.message


class Cursor(object):
    """
    Byte array cursor

    Represents a position within a byte array.

    encode_* operations write data at the current position
    and advance it by the number of bytes written.

    decode_* operations read data at the current position,
    advance it by the number of bytes read, and return
    the result.

    Cursors support in-place addition and subtraction (+=,-=) of
    integer values to manually adjust position, as well
    as ordinary addition and subtraction, which return a new
    cursor without modifiying the original.

    Ordinary subtraction of two cursors will yield the
    difference between their positions as an integer.
    ==, !=, <, <=, >, and >= operators will also work to
    compare positions.  The results will only be meaningful
    for cursors referencing the same underlying array.

    For a given cursor, cursor.hole.encode_* will perform
    the same operation as cursor.encode_*, but will return
    a hole object.  Calling this hole object with the same
    type of arguments as the original encode method will
    overwrite the original value.  This mechanism is useful
    for backpatching values into fields that aren't known until
    later, such as internal packet lengths and offsets or
    checksum fields.  For example::

        hole = cursor.hole.encode_uint32le(0)
        # Encode rest of packet
        ...
        # Overwrite value with calculated checksum
        hole(sum)

    Cursors support slicing to extract sections
    of the underlying array.  For example::

        # Extract array slice between cur1 and cur2
        subarray = cur1[:cur2]

    Cursors also support establishing boundaries outside of which
    decoding will raise exceptions::

        with cur.bounded(startcur, endcur):
            # Within this block, attempts to decode data outside of
            # the range starting with startcur (inclusive) and ending
            # with endcur (exclusive) will raise BufferOverrun().
            # If the start and end paremeters are numbers rather than
            # other cursors, they will be taken to be relative to
            # cur itself.
            ...

    @ivar array: Array referenced by cursor
    @ivar offset: Offset within the array
    @ivar bounds: Pair of lower and upper bound on offset
    """

    def __init__(self, arr, offset, bounds=(None, None)):
        """
        Create a L{Cursor} for the given array
        at the given offset.

        @type arr: array.array('B', ...)
        @param arr: The array
        @type offset: number
        @param offset: The offset from the start of the array
        @param bounds: A pair of a lower and upper bound on valid offsets
        """

        self.array = arr
        self.offset = offset
        self.bounds = bounds
        self.hole = Cursor.Hole(self)

    def __eq__(self, o):
        return self.array is o.array and self.offset == o.offset

    def __ne__(self, o):
        return not (self == o)

    def __lt__(self, o):
        assert self.array is o.array
        return self.offset < o.offset

    def __gt__(self, o):
        assert self.array is o.array
        return self.offset > o.offset

    def __le__(self, o):
        assert self.array is o.array
        return self.offset <= o.offset

    def __ge__(self, o):
        assert self.array is o.array
        return self.offset >= o.offset

    def __add__(self, o):
        return Cursor(self.array, self.offset + o, self.bounds)

    def __sub__(self, o):
        if isinstance(o, Cursor):
            assert self.array is o.array
            return self.offset - o.offset
        else:
            return Cursor(self.array, self.offset - o, self.bounds)

    def __iadd__(self, o):
        self.offset += o
        return self

    def __isub__(self, o):
        self.offset -= o
        return self

    def _getindex(self, ind):
        if ind is None:
            return None
        elif isinstance(ind, Cursor):
            assert self.array is ind.array
            return ind.offset
        else:
            return self.offset + ind

    def __getitem__(self, index):
        if isinstance(index, slice):
            start = self._getindex(index.start if index.start else 0)
            stop = self._getindex(index.stop)
            step = index.step

            self._check_bounds(start, stop)
            return self.array.__getitem__(slice(start, stop, step))
        else:
            self._check_bounds(index, index + 1)
            return self.array.__getitem__(index)

    def __repr__(self):
        return (
            "Cursor("
            + object.__repr__(self.array)
            + ","
            + repr(self.offset)
            + ","
            + repr(self.bounds)
            + ")"
        )

    def copy(self):
        """Create copy of cursor."""
        return Cursor(self.array, self.offset, self.bounds)

    def _expand_to(self, size):
        cur_size = len(self.array)
        if size > cur_size:
            self.array.extend([0] * (size - cur_size))

    def encode_bytes(self, val):
        """Encode bytes.  Accepts byte arrays, strings, and integer lists."""
        size = len(val)
        self._expand_to(self.offset + size)
        self.array[self.offset : self.offset + size] = array.array("B", val)
        self.offset += size

    def encode_struct(self, fmt, *args):
        size = struct.calcsize(fmt)
        self._expand_to(self.offset + size)
        struct.pack_into(fmt, self.array, self.offset, *args)
        self.offset += size

    def encode_uint8be(self, val):
        self.encode_struct(">B", val)

    def encode_uint16be(self, val):
        self.encode_struct(">H", val)

    def encode_uint32be(self, val):
        self.encode_struct(">L", val)

    def encode_uint64be(self, val):
        self.encode_struct(">Q", val)

    def encode_uint8le(self, val):
        self.encode_struct("<B", val)

    def encode_uint16le(self, val):
        self.encode_struct("<H", val)

    def encode_uint32le(self, val):
        self.encode_struct("<L", val)

    def encode_uint64le(self, val):
        self.encode_struct("<Q", val)

    def encode_int64le(self, val):
        self.encode_struct("<q", val)

    def encode_utf16le(self, val):
        self.encode_bytes(str(val).encode("utf-16le"))

    def trunc(self):
        self._expand_to(self.offset)
        del self.array[self.offset :]

    def _check_bounds(self, start, end):
        lower = self.bounds[0] if self.bounds[0] is not None else 0
        upper = self.bounds[1] if self.bounds[1] is not None else len(self.array)

        if start < lower:
            raise BufferOverrun(self, start, lower)
        if end > upper:
            raise BufferOverrun(self, end, upper)

    def decode_bytes(self, size):
        self._check_bounds(self.offset, self.offset + size)
        result = self.array[self.offset : self.offset + size]
        self.offset += size
        return result

    def decode_struct(self, fmt):
        size = struct.calcsize(fmt)
        self._check_bounds(self.offset, self.offset + size)
        result = struct.unpack_from(fmt, self.array, self.offset)
        self.offset += size
        return result

    def decode_uint8be(self):
        return self.decode_struct(">B")[0]

    def decode_uint16be(self):
        return self.decode_struct(">H")[0]

    def decode_uint32be(self):
        return self.decode_struct(">L")[0]

    def decode_uint64be(self):
        return self.decode_struct(">Q")[0]

    def decode_uint8le(self):
        return self.decode_struct("<B")[0]

    def decode_uint16le(self):
        return self.decode_struct("<H")[0]

    def decode_uint32le(self):
        return self.decode_struct("<L")[0]

    def decode_int32le(self):
        return self.decode_struct("<l")[0]

    def decode_uint64le(self):
        return self.decode_struct("<Q")[0]

    def decode_int64le(self):
        return self.decode_struct("<q")[0]

    def decode_utf16le(self, size):
        return self.decode_bytes(size).tobytes().decode("utf-16le")

    def align(self, base, val):
        assert self.array is base.array
        rem = (self.offset - base.offset) % val
        if rem != 0:
            self.offset += val - rem

    def seekto(self, o, lowerbound=None, upperbound=None):
        assert self.array is o.array
        if lowerbound is not None and o < lowerbound:
            raise BufferOverrun(self, o, lowerbound)
        if upperbound is not None and o > upperbound:
            raise BufferOverrun(self, o, upperbound)
        self.offset = o.offset

    def advanceto(self, o, bound=None):
        self.seekto(o, self, bound)

    def reverseto(self, o, bound=None):
        self.seekto(o, bound, self)

    @property
    def lowerbound(self):
        lower = self.bounds[0] if self.bounds[0] is not None else 0
        return Cursor(self.array, lower, self.bounds)

    @property
    def upperbound(self):
        upper = self.bounds[1] if self.bounds[1] is not None else len(self.array)
        return Cursor(self.array, upper, self.bounds)

    def bounded(self, lower, upper):
        # Allow cursors to be used as bounds (the preferred idiom)
        if isinstance(lower, Cursor):
            assert self.array is lower.array
            lower = lower.offset
        else:
            lower = cur.offset + lower
        if isinstance(upper, Cursor):
            assert self.array is upper.array
            upper = upper.offset
        else:
            upper = cur.offset + upper

        # Don't let new bounds escape current bounds
        self._check_bounds(lower, upper)

        return Cursor.Bounds(self, lower, upper)

    class Hole(object):
        def __init__(self, cur):
            self.cur = cur

        def __getattr__(self, attr):
            if hasattr(self.cur.__class__, attr):
                f = getattr(self.cur.__class__, attr)
                if callable(f):
                    copy = self.cur.copy()

                    def f2(*args, **kwargs):
                        offset = copy.offset
                        f(copy, *args, **kwargs)
                        copy.offset = offset

                    def f1(*args, **kwargs):
                        f(self.cur, *args, **kwargs)
                        return f2

                    return f1
                else:
                    raise AttributeError
            else:
                raise AttributeError

    class Bounds(object):
        def __init__(self, cur, lower, upper):
            self.cur = cur
            self.bounds = (lower, upper)

        def __enter__(self):
            self.oldbounds = self.cur.bounds
            self.cur.bounds = self.bounds
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.cur.bounds = self.oldbounds


class BadPacket(Exception):
    pass


class FrameMeta(type):
    def __new__(mcs, name, bases, dict):
        # Inherit _register from bases
        dict["_register"] = []
        for base in bases:
            if hasattr(base, "_register"):
                dict["_register"] += base._register

        # Inherit field_blacklist from bases
        if "field_blacklist" in dict:
            for base in bases:
                if hasattr(base, "field_blacklist"):
                    dict["field_blacklist"] += base.field_blacklist

        result = type.__new__(mcs, name, bases, dict)

        # Register class in appropriate tables
        for (table, keyattrs) in result._register:
            if all(hasattr(result, a) for a in keyattrs):
                key = [getattr(result, a) for a in keyattrs]
                if len(key) == 1:
                    key = key[0]
                else:
                    key = tuple(key)
                table[key] = result

        return result


class Frame(with_metaclass(FrameMeta)):
    field_blacklist = ["fields", "parent", "start", "end"]
    LOG_CHILDREN_COUNT = False  # Include len(children) in __repr__
    LOG_CHILDREN_EXPAND = False  # Include c._log_str for all children

    def __init__(self, parent, context=None):
        object.__setattr__(self, "fields", [])
        self.parent = parent
        self._context = context

    def __len__(self):
        return len(self.children)

    def __getitem__(self, key):
        return self.children[key]

    def __iter__(self):
        return self.children.__iter__()

    def __setattr__(self, name, value):
        if (
            not name.startswith("_")
            and name not in self.fields
            and name not in self.field_blacklist
        ):
            self.fields.append(name)
        object.__setattr__(self, name, value)

    def __str__(self):
        return self._str(1)

    @staticmethod
    def _value_str(value):
        if isinstance(value, array.array) and value.typecode == "B":
            return "0x" + "".join("%.2x" % b for b in value)
        else:
            return str(value)

    def _str(self, indent):
        res = self.__class__.__name__
        for field in self.fields:
            value = getattr(self, field)
            if value is not None:
                if isinstance(value, Frame):
                    valstr = value._str(indent + 1)
                else:
                    valstr = self._value_str(value)
                res += "\n" + "  " * indent + field + ": " + valstr
        for child in self.children:
            valstr = child._str(indent + 1)
            res += "\n" + "  " * indent + valstr
        return res

    def _log_str_expand_children(self):
        """
        return a list of _log_str output for all children
        """
        return [c._log_str() for c in self.children]

    def _log_str_count_children(self):
        """
        return a list containing a string count of children wrapped in
        parentheses
        """
        if self.children:
            return ["({})".format(len(self.children))]
        return []

    def _log_str_children(self):
        """
        return a list of string components summarizing the children
        """
        components = []
        if self.LOG_CHILDREN_COUNT:
            components.extend(self._log_str_count_children())
        if self.LOG_CHILDREN_EXPAND:
            components.extend(self._log_str_expand_children())
        return components

    def _log_str(self):
        """
        Short description used for logging. Should provide basic context and be
        under 100 characters
        """
        components = [type(self).__name__]
        components.extend(self._log_str_children())
        return " ".join(components)

    def _encode_pre(self, cur):
        self.start = cur.copy()

    def _encode_post(self, cur):
        self.end = cur.copy()

    def _decode_pre(self, cur):
        self.start = cur.copy()

    def _decode_post(self, cur):
        self.end = cur.copy()

    @property
    def context(self):
        if self._context is not None:
            return self._context
        elif self.parent is not None:
            return self.parent.context
        else:
            return None

    @property
    def children(self):
        return self._children() if hasattr(self, "_children") else []

    def encode(self, cur):
        self._encode_pre(cur)
        self._encode(cur)
        self._encode_post(cur)

    def decode(self, cur):
        self._decode_pre(cur)
        self._decode(cur)
        self._decode_post(cur)

    def serialize(self):
        self.buf = array.array("B")
        cursor = Cursor(self.buf, 0)
        self.encode(cursor)
        return self.buf

    def parse(self, arr):
        cursor = Cursor(arr, 0)
        self.decode(cursor)
        self.buf = arr

    def next_sibling(self):
        children = self.parent.children
        index = children.index(self) + 1
        return children[index] if index < len(children) else None

    def prev_sibling(self):
        children = self.parent.children
        index = children.index(self) - 1
        return children[index] if index >= 0 else None

    def is_last_child(self):
        children = self.parent.children
        return children.index(self) == len(children) - 1


class Register(object):
    def __init__(self, table, *keyattrs):
        self.table = table
        self.keyattrs = keyattrs

    def __call__(self, cls):
        assert issubclass(cls, Frame)
        cls._register.append((self.table, self.keyattrs))
        return cls


class EnumMeta(type):
    def __new__(mcs, cname, bases, idict):
        nametoval = {}
        valtoname = {}
        misc = {}

        for (name, val) in idict.items():
            if name[0].isupper():
                nametoval[name] = val
                valtoname[val] = name
            else:
                misc[name] = val

        cls = type.__new__(mcs, cname, bases, misc)
        cls._nametoval = nametoval
        cls._valtoname = valtoname

        return cls

    def __getattribute__(cls, name):
        nametoval = type.__getattribute__(cls, "_nametoval")
        if name in nametoval:
            return cls(nametoval[name])
        else:
            return type.__getattribute__(cls, name)


class Enum(with_metaclass(EnumMeta, int)):
    """
    Enumeration abstract base

    An Enum subclasses long in order to ensure that instances
    are a member of a well-defined enumeration of values,
    provide introspection into the allowed values and their names,
    and provide symbolic string forms of values.

    You should generally subclass one of L{ValueEnum} or L{FlagEnum}.
    """

    @classmethod
    def items(cls):
        """
        Returns a list of (name,value) pairs for allowed enumeration values.
        """
        return iter(cls._nametoval.items())

    @classmethod
    def names(cls):
        """
        Returns a list of names of allowed enumeration values.
        """
        return list(cls._nametoval.keys())

    @classmethod
    def values(cls):
        """
        Returns a list of allowed enumeration values.
        """
        return list(cls._nametoval.values())

    @classmethod
    def import_items(cls, dictionary):
        """
        Import enumeration values into dictionary

        This method is intended to allow importing enumeration
        values from an L{Enum} subclass into the root of a module,
        and should be invoked as::

            SomeEnumClass.import_items(globals())
        """
        dictionary.update((name, cls(value)) for (name, value) in cls.items())

    @classmethod
    def validate(cls, value):
        """
        Validate value as valid for enumeration.

        This must be implemented in subclasses of L{Enum}
        (but not L{ValueEnum} or L{FlagEnum}, which provide it).
        It should raise a ValueError on failure.
        """
        raise NotImplementedError()

    def __new__(cls, value=0):
        """
        Constructor.

        Creates a new Enum instance from an ordinary number,
        validating that is is valid for the particular
        enumeration.
        """
        cls.validate(value)
        return super(Enum, cls).__new__(cls, value)

    def __repr__(self):
        # Just return string form
        return str(self)


class ValueEnum(Enum):
    """
    Value Enumeration

    Subclass this class to define enumerations of values.
    For example::

        class SomeEnumClass(Enum):
            FOO = 0
            BAR = 1
            BAZ = 2

    Accessing SomeEnumClass.FOO will actually return a SomeEnumClass instance.
    Instances will return their symbolic names when str() is used.
    """

    permissive = False

    @classmethod
    def validate(cls, value):
        if not cls.permissive and value not in cls._valtoname:
            raise ValueError("Invalid %s: %x" % (cls.__name__, value))

    def __str__(self):
        if self in self.__class__._valtoname:
            return self.__class__._valtoname[self]
        else:
            return hex(self)


class FlagEnum(Enum):
    """
    Flag Enumeration

    Subclass this class to define enumerations of flags.
    For example::

        class SomeFlagEnumClass(Enum):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4

    Accessing SomeFlagEnumClass.FOO will actually return a
    SomeFlagEnumClass instance.  Instances may be bitwise-ored
    together.  Instances will return the symbolic names of all
    set flags joined by ' | ' when str() is used.
    """

    @classmethod
    def validate(cls, value):
        remaining = value
        for flag in cls.values():
            if flag & remaining == flag:
                remaining &= ~flag

        if remaining != 0:
            raise ValueError(
                "Invalid %s: 0x%x (remainder 0x%x)" % (cls.__name__, value, remaining)
            )

    def __str__(self):
        names = [
            name
            for (name, flag) in self.items()
            if (flag != 0 and flag & self == flag) or (self == 0 and flag == 0)
        ]

        return " | ".join(names) if len(names) else "0"

    def __or__(self, o):
        return self.__class__(super(FlagEnum, self).__or__(o))

    def __and__(self, o):
        return self.__class__(super(FlagEnum, self).__and__(o))


class Let(object):
    """
    A Let allows for temporarily storing passed arguments in the _settings
    member of a target object for the duration of the contextmanager's scope

    .. code-block:: python
        :caption: Implementation on a factory class

        class MyThingFactory(object):
            def __init__(self):
                self._settings = {}
            def generate_thing(self):
                # do something here
                new_obj = Thing()
                for k,v in self._settings.iteritems():
                    setattr(new_obj, k, v)
                return new_obj
            def let(self, **kwds):
                return Let(self, kwds)

    Calling code that wants to temporarily customize the ``Thing`` objects that
    are returned by the factory:

    .. code-block:: python

        fac = MyThingFactory()
        with fac.let(this="that", foo="bar"):
            a_thing = fac.generate_thing()
            self.assertEqual(a_thing.this, "that")
    """

    def __init__(self, target, settings_dict):
        self.target = target
        self.settings_dict = settings_dict
        if not hasattr(target, "_settings"):
            target._settings = {}

    def __enter__(self):
        self.backup = dict(self.target._settings)
        self.target._settings.update(self.settings_dict)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.target._settings = self.backup



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        changenotify.py
#
# Abstract:
#
#        Change Notify tests
#
# Authors: Masen Furer <masen.furer@emc.com>
#

import pike.model
import pike.ntstatus
import pike.smb2
import pike.test


class ChangeNotifyTest(pike.test.PikeTest):
    def test_change_notify_file_name(self):
        filename = "change_notify.txt"
        chan, tree = self.tree_connect()

        # connect to the root of the share
        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()

        # build a change notify request
        smb_req = chan.request(obj=root)
        notify_req = pike.smb2.ChangeNotifyRequest(smb_req)
        notify_req.file_id = root.file_id
        notify_req.completion_filter = pike.smb2.SMB2_NOTIFY_CHANGE_FILE_NAME
        # send the request async
        futures = chan.connection.submit(smb_req.parent)

        # create a file on the share to trigger the notification
        file = chan.create(
            tree,
            filename,
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()

        chan.close(file)

        # collect the change notify response
        result = futures[0].result()[0]

        # expect one notification, for the file add
        self.assertEqual(len(result.notifications), 1)
        self.assertEqual(result.notifications[0].filename, filename)
        self.assertEqual(result.notifications[0].action, pike.smb2.SMB2_ACTION_ADDED)

        chan.close(root)

    def test_change_notify_cancel(self):
        chan, tree = self.tree_connect()

        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()
        cn_future = chan.change_notify(root)
        # close the handle for the outstanding notify
        chan.close(root)

        smb2_resp = cn_future.result()
        # [MS-SMB2] 3.3.5.10 - The Server MUST send an SMB2 CHANGE_NOTIFY Response with
        # STATUS_NOTIFY_CLEANUP status code for all pending CHANGE_NOTIFY requests
        # associated with the FileId that is closed.
        self.assertEqual(smb2_resp.status, pike.ntstatus.STATUS_NOTIFY_CLEANUP)
        notify_resp = smb2_resp[0]
        # The response should contain no FileNotifyInformation structs
        self.assertEqual(len(notify_resp), 0)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        lock.py
#
# Abstract:
#
#        Lock tests
#
# Authors: Arlene Berry (arlene.berry@emc.com)
#

from builtins import range
import pike.model
import pike.smb2
import pike.test
import pike.ntstatus


class LockTest(pike.test.PikeTest):
    # Take a basic byte-range lock
    def test_lock(self):
        chan, tree = self.tree_connect()
        buffer = "0123456789012345678901"
        locks = [
            (
                8,
                8,
                pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK
                | pike.smb2.SMB2_LOCKFLAG_FAIL_IMMEDIATELY,
            )
        ]

        share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )

        file = chan.create(
            tree,
            "lock.txt",
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()

        bytes_written = chan.write(file, 0, buffer)
        self.assertEqual(bytes_written, len(buffer))

        chan.lock(file, locks).result()

        chan.close(file)

    # Test that pending lock request can be cancelled, yielding STATUS_CANCELLED
    def test_cancel(self):
        chan, tree = self.tree_connect()
        buffer = "0123456789012345678901"
        locks = [(8, 8, pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK)]

        share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )
        access = pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE

        # Create file, lock
        file1 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()

        bytes_written = chan.write(file1, 0, buffer)
        self.assertEqual(bytes_written, len(buffer))

        chan.lock(file1, locks).result()

        # Open file again (with delete on close)
        file2 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_OPEN,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()

        # This will block since the lock is already held, so only wait for interim response
        lock_future = chan.lock(file2, locks)
        lock_future.wait_interim()

        # Cancel, wait for response, verify error response
        with self.assert_error(pike.ntstatus.STATUS_CANCELLED):
            chan.cancel(lock_future).result()

        chan.close(file1)
        chan.close(file2)

    def test_deny_write(self):
        chan, tree = self.tree_connect()
        buffer = "0123456789012345678901"
        lock_offset = 8
        lock_size = 8
        locks = [(lock_offset, lock_size, pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK)]

        share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )
        access = pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE

        # Create file, lock
        file1 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()

        bytes_written = chan.write(file1, 0, buffer)
        self.assertEqual(bytes_written, len(buffer))

        chan.lock(file1, locks).result()

        # Open file again (with delete on close)
        file2 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_OPEN,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()

        cases = ((offset, size) for offset in range(0, 16) for size in range(0, 16))

        for (offset, size) in cases:
            if ranges_intersect(
                offset, offset + size, lock_offset, lock_offset + lock_size
            ):
                with self.assert_error(pike.ntstatus.STATUS_FILE_LOCK_CONFLICT):
                    chan.write(file2, offset, "a" * size)

        chan.close(file1)
        chan.close(file2)

    def test_allow_zero_byte_write(self):
        chan, tree = self.tree_connect()
        buffer = "0123456789012345678901"
        lock_offset = 8
        lock_size = 8
        locks = [(lock_offset, lock_size, pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK)]

        share_all = (
            pike.smb2.FILE_SHARE_READ
            | pike.smb2.FILE_SHARE_WRITE
            | pike.smb2.FILE_SHARE_DELETE
        )
        access = pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE

        # Create file, lock
        file1 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()

        bytes_written = chan.write(file1, 0, buffer)
        self.assertEqual(bytes_written, len(buffer))

        chan.lock(file1, locks).result()

        # Open file again (with delete on close)
        file2 = chan.create(
            tree,
            "lock.txt",
            access=access,
            share=share_all,
            disposition=pike.smb2.FILE_OPEN,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()

        chan.write(file2, lock_offset + 1, None)

        chan.close(file1)
        chan.close(file2)


def range_contains(a1, a2, b):
    return b >= a1 and b < a2


def ranges_intersect(a1, a2, b1, b2):
    return (
        a2 > a1
        and b2 > b1
        and (
            range_contains(a1, a2, b1)
            or range_contains(a1, a2, b2 - 1)
            or range_contains(b1, b2, a1)
            or range_contains(b1, b2, a2 - 1)
        )
    )



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        querydirectory.py
#
# Abstract:
#
#        Query directory tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.model
import pike.smb2
import pike.test
import pike.ntstatus


class QueryDirectoryTest(pike.test.PikeTest):
    # Enumerate directory at FILE_DIRECTORY_INFORMATION level.
    # Test for presence of . and .. entries
    def test_file_directory_info(self):

        chan, tree = self.tree_connect()
        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()

        names = [info.file_name for info in chan.enum_directory(root)]

        self.assertIn(".", names)

        chan.close(root)

    # Querying for a specific filename twice
    # on the same handle succeeds the first time and
    # fails with STATUS_NO_MORE_FILES the second.
    def test_specific_name(self):
        chan, tree = self.tree_connect()
        name = "hello.txt"

        hello = chan.create(
            tree,
            name,
            access=pike.smb2.GENERIC_WRITE | pike.smb2.GENERIC_READ | pike.smb2.DELETE,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()

        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()

        query1 = chan.query_directory(root, file_name=name)

        with self.assert_error(pike.ntstatus.STATUS_NO_MORE_FILES):
            chan.query_directory(root, file_name=name)

        chan.close(hello)
        chan.close(root)

    def test_file_id_both_directory_information(self):

        chan, tree = self.tree_connect()
        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()

        result = chan.query_directory(
            root, file_information_class=pike.smb2.FILE_ID_BOTH_DIR_INFORMATION
        )
        names = [info.file_name for info in result]
        self.assertIn(".", names)
        self.assertIn("..", names)

        valid_file_ids = [info.file_id >= 0 for info in result]
        self.assertNotIn(False, valid_file_ids)

        chan.close(root)

    def test_restart_scan(self):

        chan, tree = self.tree_connect()
        root = chan.create(
            tree,
            "",
            access=pike.smb2.GENERIC_READ,
            options=pike.smb2.FILE_DIRECTORY_FILE,
            share=pike.smb2.FILE_SHARE_READ,
        ).result()

        result = chan.query_directory(root)
        names = [info.file_name for info in result]
        self.assertIn(".", names)

        result = chan.query_directory(
            root, flags=pike.smb2.SL_RESTART_SCAN, file_name="*"
        )
        names = [info.file_name for info in result]
        self.assertIn(".", names)

        chan.close(root)



"""
Exceptions used by multiple modules are defined here
"""

from . import core


class TimeoutError(Exception):
    """Future completion timed out"""

    future = None

    @classmethod
    def with_future(cls, future, *args):
        """
        Instantiate TimeoutError from a given future.

        :param future: Future that timed out
        :param args: passed to Exception.__init__
        :return: TimeoutError
        """
        ex = cls(*args)
        ex.future = future
        return ex

    def __str__(self):
        s = super(TimeoutError, self).__str__()
        if self.future is not None:
            if self.future.request is not None:
                requests = [str(self.future.request)]
                if not isinstance(self.future.request, (core.Frame, str, bytes)):
                    # attempt to recursively str format other iterables
                    try:
                        requests = [str(r) for r in self.future.request]
                    except TypeError:
                        pass
                s += "\nRequest: {}".format("\n".join(requests))
            if self.future.interim_response is not None:
                s += "\nInterim: {}".format(self.future.interim_response)
        return s


class StateError(Exception):
    pass


class CreditError(Exception):
    pass


class RequestError(Exception):
    def __init__(self, request, message=None):
        if message is None:
            message = "Could not send {0}".format(repr(request))
        Exception.__init__(self, message)
        self.request = request


class CallbackError(Exception):
    """
    the callback was not suitable
    """


class ResponseError(Exception):
    """
    Raised when an Smb2 response contains an unexpected NTSTATUS.
    """

    def __init__(self, response):
        Exception.__init__(self, response.command, response.status)
        self.response = response



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        session.py
#
# Abstract:
#
#        Session setup tests
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

import pike.test


class SessionTest(pike.test.PikeTest):
    # Log off a session
    def test_session_logoff(self):
        chan, tree = self.tree_connect()
        chan.logoff()

    def test_session_multiplex(self):
        chan, tree = self.tree_connect()
        chan2 = chan.connection.session_setup(self.creds)
        chan3 = chan.connection.session_setup(self.creds)
        self.assertEqual(chan.connection, chan2.connection)
        self.assertEqual(chan2.connection, chan3.connection)
        self.assertNotEqual(chan.session, chan2.session)
        self.assertNotEqual(chan2.session, chan3.session)
        self.assertNotEqual(chan.session.session_id, chan2.session.session_id)
        self.assertNotEqual(chan2.session.session_id, chan3.session.session_id)
        self.assertNotEqual(chan.session.session_key, chan2.session.session_key)
        self.assertNotEqual(chan2.session.session_key, chan3.session.session_key)
        tree2 = chan2.tree_connect(self.share)
        tree3 = chan3.tree_connect(self.share)
        chan3.logoff()
        chan2.logoff()
        chan.logoff()



#
# Copyright (c) 2015-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        copychunk.py
#
# Abstract:
#
#        Tests for server side copy
#
# Authors: Avi Bhandari (avi.bahndari@emc.com)
#          Masen Furer (masen.furer@emc.com)
#

from __future__ import division
from builtins import chr
from builtins import range
from builtins import str
from builtins import zip

import array
import random

import pike.ntstatus
import pike.smb2
import pike.test

share_all = (
    pike.smb2.FILE_SHARE_READ | pike.smb2.FILE_SHARE_WRITE | pike.smb2.FILE_SHARE_DELETE
)
access_rwd = pike.smb2.FILE_READ_DATA | pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE
access_wd = pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE

SIMPLE_CONTENT = "Hello"
SIMPLE_5_CHUNKS = [
    (0, 0, 4000),
    (4000, 4000, 4000),
    (8000, 8000, 4000),
    (12000, 12000, 4000),
    (16000, 16000, 4000),
]
SIMPLE_5_CHUNKS_LEN = 20000


def _gen_test_buffer(length):
    # XXX: refactor and push up into helper/util module
    pattern = "".join([chr(x) for x in range(ord(" "), ord("~"))])
    buf = (pattern * ((length // len(pattern)) + 1))[:length]
    return buf.encode("ascii")


def _gen_random_test_buffer(length):
    # XXX: refactor and push up into helper/util module
    buf = ""
    random_str_seq = "0123456789" + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    for i in range(0, length):
        if i % length == 0 and i != 0:
            buf += "-"
        # XXX: accidentally quadratic str concat
        buf += str(random_str_seq[random.randint(0, len(random_str_seq) - 1)])
    return buf.encode("ascii")


###
# Main test
###


class TestServerSideCopy(pike.test.PikeTest):
    # these values are valid for Windows -- other servers may need to adjust
    bad_resume_key_error = pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND
    max_number_of_chunks = 256
    max_chunk_size = 1048576
    max_data_size = 16777216

    def setUp(self):
        self.chan, self.tree = self.tree_connect()
        self.other_chan_list = []

    def tearDown(self):
        self._clean_up_other_channels()
        self.chan.tree_disconnect(self.tree)
        self.chan.logoff()

    def _create_and_write(self, filename, content):
        """create / overwrite filename with content"""
        fh1 = self.chan.create(
            self.tree,
            filename,
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
        ).result()
        # get the max size of write per request
        max_write_size = min(
            self.chan.connection.negotiate_response.max_write_size, self.max_chunk_size
        )
        total_len = len(content)
        this_offset = 0
        writes = []
        while this_offset < total_len:
            offset = this_offset
            if this_offset + max_write_size < total_len:
                length = max_write_size
            else:
                length = total_len - this_offset
            writes.append((offset, length))
            this_offset += max_write_size
        num_of_writes = len(writes)
        bytes_written = 0
        for the_offset, the_length in writes:
            bytes_written += self.chan.write(
                fh1, the_offset, content[the_offset : (the_offset + the_length)]
            )
        self.chan.close(fh1)

    def _open_src_dst(
        self,
        src_filename,
        dst_filename,
        src_access=None,
        src_disp=None,
        src_options=None,
        dst_access=None,
        dst_disp=None,
        dst_options=None,
    ):
        if src_access is None:
            src_access = pike.smb2.FILE_READ_DATA | pike.smb2.DELETE
        if dst_access is None:
            dst_access = access_rwd
        if src_disp is None:
            src_disp = pike.smb2.FILE_OPEN
        if dst_disp is None:
            dst_disp = pike.smb2.FILE_SUPERSEDE
        if src_options is None:
            src_options = pike.smb2.FILE_DELETE_ON_CLOSE
        if dst_options is None:
            dst_options = pike.smb2.FILE_DELETE_ON_CLOSE

        if src_filename == dst_filename:
            src_options = 0  # if we're opening the same file, only delete one
        fh_src = self.chan.create(
            self.tree,
            src_filename,
            access=src_access,
            share=share_all,
            disposition=src_disp,
            options=src_options,
        ).result()

        fh_dst = self.chan.create(
            self.tree,
            dst_filename,
            access=dst_access,
            share=share_all,
            disposition=dst_disp,
            options=dst_options,
        ).result()
        return (fh_src, fh_dst)

    def _get_readable_handler(self, filename):
        fh_read = self.chan.create(
            self.tree,
            filename,
            access=pike.smb2.FILE_READ_DATA,
            share=share_all,
            options=0,
        ).result()
        return fh_read

    def _clean_up_other_channels(self):
        """
        clean up chan tree filehandles during failure
        """
        if self.other_chan_list:
            for info in self.other_chan_list:
                for fh in info["filehandles"]:
                    info["channel"].close(fh)
                info["channel"].tree_disconnect(info["tree"])
                info["channel"].logoff()
            self.other_chan_list = []

    def _prepare_source_file(self):
        """
        generate a five chunks file as source file for resume_key request
        """
        src_filename = "src_copy_chunk_rk.txt"
        self._create_and_write(src_filename, SIMPLE_CONTENT)

    def _read_big_file(self, file_handle, total_len, add_offset=0):
        """read large file and return in string"""
        max_read_size = self.chan.connection.negotiate_response.max_read_size
        reads = []
        this_offset = 0
        while this_offset < total_len:
            offset = this_offset
            if this_offset + max_read_size < total_len:
                length = max_read_size
            else:
                length = total_len - this_offset
            reads.append((offset, length))
            this_offset += max_read_size
        num_of_reads = len(reads)
        read_list = []
        for the_offset, the_length in reads:
            read_list.append(
                self.chan.read(
                    file_handle, the_length, the_offset + add_offset
                ).tobytes()
            )
        return b"".join(read_list)

    def _gen_16mega_file(self, filename):
        """
        use multiple server side copy to generate a 16M sourcefile
        """
        file_64k = "temp_src_file_64k.txt"
        file_1m = "temp_src_file_1m.txt"
        block_64k = _gen_test_buffer(65536)
        self._create_and_write(file_64k, block_64k)
        list_offset_64k = [i * 65536 for i in range(16)]
        chunks_64k_1m = list(zip([0] * 16, list_offset_64k, [65536] * 16))
        list_offset_1m = [i * 1048576 for i in range(16)]
        chunks_1m_16m = list(zip([0] * 16, list_offset_1m, [1048576] * 16))
        fh_64k = self.chan.create(
            self.tree,
            file_64k,
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_OPEN,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()
        fh_1m = self.chan.create(
            self.tree,
            file_1m,
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
        ).result()
        fh_16m = self.chan.create(
            self.tree,
            filename,
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=0,
        ).result()
        self.chan.copychunk(fh_64k, fh_1m, chunks_64k_1m)
        self.chan.copychunk(fh_1m, fh_16m, chunks_1m_16m)
        self.chan.close(fh_64k)
        self.chan.close(fh_1m)
        self.chan.close(fh_16m)

    def _gen_random_chunks_on_para(self, random_chunk_size=False, random_offset=False):
        """
        generate random chunks for chunkcopy operation
        copy length is always randomized, chunk_size and offset can be fixed by flags
        """

        if random_chunk_size:
            chunk_sz = random.randrange(1, self.max_chunk_size + 1)
        else:
            chunk_sz = self.max_chunk_size
        total_len = random.randrange(1, 16 * chunk_sz)
        if random_offset:
            dst_offset = random.randrange(1, 4294967295 - total_len)
        else:
            dst_offset = 0
        this_offset = 0
        chunks = []
        while this_offset < total_len:
            offset = this_offset
            if this_offset + chunk_sz < total_len:
                length = chunk_sz
            else:
                length = total_len - this_offset
            chunks.append((offset, offset + dst_offset, length))
            this_offset += chunk_sz
        self.logger.info(
            "chunks generated, total_length is %d, chunk_size is %d, dst_offset is %lu"
            % (total_len, chunk_sz, dst_offset)
        )
        return chunks

    def _prepare_before_ssc_copy(self, filepair, write_thru=False):
        if write_thru:
            my_options = pike.smb2.FILE_WRITE_THROUGH
        else:
            my_options = 0
        num_sess = len(filepair)
        for i in range(0, num_sess):
            chan, tree = self.tree_connect()
            fh_src = chan.create(
                tree, filepair[i][0], access=access_rwd, share=share_all
            ).result()

            fh_dst = chan.create(
                tree,
                filepair[i][1],
                access=access_rwd,
                share=share_all,
                options=my_options,
            ).result()
            resume_key = chan.resume_key(fh_src)[0][0].resume_key
            self.other_chan_list.append(
                {
                    "channel": chan,
                    "tree": tree,
                    "filehandles": [fh_src, fh_dst],
                    "resume_key": resume_key,
                }
            )

    def _submit_ssc_async_request(self, chan, file_handle, resume_key, chunks):
        """
        submit ssc request with resume_key as input parameter
        return the future object for later check
        """
        smb_req = chan.request(obj=file_handle.tree)
        ioctl_req = pike.smb2.IoctlRequest(smb_req)
        copychunk_req = pike.smb2.CopyChunkCopyRequest(ioctl_req)

        ioctl_req.max_output_response = 16384
        ioctl_req.file_id = file_handle.file_id
        ioctl_req.flags |= pike.smb2.SMB2_0_IOCTL_IS_FSCTL
        copychunk_req.source_key = resume_key
        copychunk_req.chunk_count = len(chunks)

        for source_offset, target_offset, length in chunks:
            chunk = pike.smb2.CopyChunk(copychunk_req)
            chunk.source_offset = source_offset
            chunk.target_offset = target_offset
            chunk.length = length

        future = chan.connection.submit(ioctl_req.parent.parent)
        return future

    def _submit_ssc_async_request_in_other_channels(self, index, chunks):
        chan = self.other_chan_list[index]["channel"]
        fh_dst = self.other_chan_list[index]["filehandles"][1]
        resume_key = self.other_chan_list[index]["resume_key"]

        return self._submit_ssc_async_request(chan, fh_dst, resume_key, chunks)

    def generic_ssc_test_case(
        self, block, number_of_chunks, total_offset=0, write_flag=False
    ):
        """
        copy block in number_of_chunks, offset the destination copy by total_offset
        """
        src_filename = "src_copy_chunk_offset.txt"
        dst_filename = "dst_copy_chunk_offset.txt"
        self._create_and_write(src_filename, block)

        total_len = len(block)
        chunk_sz = (total_len // number_of_chunks) + 1
        this_offset = 0

        chunks = []
        while this_offset < total_len:
            offset = this_offset
            if this_offset + chunk_sz < total_len:
                length = chunk_sz
            else:
                length = total_len - this_offset
            chunks.append((offset, offset + total_offset, length))
            this_offset += chunk_sz

        if write_flag:
            dst_access = access_wd
        else:
            dst_access = access_rwd

        fh_src, fh_dst = self._open_src_dst(
            src_filename, dst_filename, dst_access=dst_access
        )

        result = self.chan.copychunk(fh_src, fh_dst, chunks, write_flag=write_flag)

        self.assertEqual(result[0][0].chunks_written, number_of_chunks)
        self.assertEqual(result[0][0].total_bytes_written, total_len)

        # read each file and verify the result
        src_buf = self.chan.read(fh_src, total_len, 0).tobytes()
        self.assertBufferEqual(src_buf, block)
        fh_dst_read = self._get_readable_handler(dst_filename)
        dst_buf = self.chan.read(fh_dst_read, total_len, total_offset).tobytes()
        self.assertBufferEqual(dst_buf, block)

        self.chan.close(fh_dst_read)
        self.chan.close(fh_src)
        self.chan.close(fh_dst)

    def test_copy_small_file(self):
        block = b"Hello"
        num_of_chunks = 1
        self.generic_ssc_test_case(block, num_of_chunks)

    def test_copy_big_file(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 1
        self.generic_ssc_test_case(block, num_of_chunks)

    def test_copy_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        self.generic_ssc_test_case(block, num_of_chunks)

    def test_copy_max_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 16
        self.generic_ssc_test_case(block, num_of_chunks)

    def test_offset_copy_small_file(self):
        block = b"Hello"
        num_of_chunks = 1
        offset = 64
        self.generic_ssc_test_case(block, num_of_chunks, offset)

    def test_offset_copy_big_file(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 1
        offset = 64
        self.generic_ssc_test_case(block, num_of_chunks, offset)

    def test_offset_copy_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        offset = 64
        self.generic_ssc_test_case(block, num_of_chunks, offset)

    def test_copy_write_small_file(self):
        block = b"Hello"
        num_of_chunks = 1
        self.generic_ssc_test_case(block, num_of_chunks, write_flag=True)

    def test_copy_write_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        self.generic_ssc_test_case(block, num_of_chunks, write_flag=True)

    def test_copy_write_max_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 16
        self.generic_ssc_test_case(block, num_of_chunks, write_flag=True)

    def generic_ssc_same_file_test_case(self, block, number_of_chunks, total_offset=0):
        """
        duplicate block in number_of_chunks to the same file,
        the copy will be total_offset from the current end of file
        """
        filename = "src_copy_chunk_same.txt"
        self._create_and_write(filename, block)

        total_len = len(block)
        chunk_sz = (total_len // number_of_chunks) + 1
        this_offset = 0

        chunks = []
        while this_offset < total_len:
            offset = this_offset
            if this_offset + chunk_sz < total_len:
                length = chunk_sz
            else:
                length = total_len - this_offset
            chunks.append((offset, offset + total_len + total_offset, length))
            this_offset += chunk_sz

        fh_src, fh_dst = self._open_src_dst(
            filename, filename, dst_disp=pike.smb2.FILE_OPEN_IF
        )

        result = self.chan.copychunk(fh_src, fh_dst, chunks)
        self.assertEqual(result[0][0].chunks_written, number_of_chunks)
        self.assertEqual(result[0][0].total_bytes_written, total_len)

        # read the file and verify the result
        src_buf = self.chan.read(fh_src, total_len, 0).tobytes()
        self.assertBufferEqual(src_buf, block)
        if total_offset > 0:
            offs_buf = self.chan.read(fh_src, total_offset, total_len).tobytes()
            self.assertBufferEqual(offs_buf, b"\x00" * total_offset)
        dst_buf = self.chan.read(fh_src, total_len, total_len + total_offset).tobytes()
        self.assertBufferEqual(dst_buf, block)

        self.chan.close(fh_src)
        self.chan.close(fh_dst)

    def test_same_small_file(self):
        block = b"Hello"
        num_of_chunks = 1
        self.generic_ssc_same_file_test_case(block, num_of_chunks)

    def test_same_big_file(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 1
        self.generic_ssc_same_file_test_case(block, num_of_chunks)

    def test_same_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        self.generic_ssc_same_file_test_case(block, num_of_chunks)

    def test_same_offset_small_file(self):
        block = b"Hello"
        num_of_chunks = 1
        offset = 64
        self.generic_ssc_same_file_test_case(block, num_of_chunks, offset)

    def test_same_offset_big_file(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 1
        offset = 64
        self.generic_ssc_same_file_test_case(block, num_of_chunks, offset)

    def test_same_offset_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        offset = 64
        self.generic_ssc_same_file_test_case(block, num_of_chunks, offset)

    def generic_ssc_overlap_test_case(
        self, block, number_of_chunks, overlap_each_block=0
    ):
        """
        copy block in number_of_chunks, each destination block offset will be overlapped
        over the previous block by overlap_each_block bytes
        """
        src_filename = "src_copy_chunk_overlap.txt"
        dst_filename = "dst_copy_chunk_overlap.txt"
        self._create_and_write(src_filename, block)

        total_len = len(block)
        chunk_sz = (total_len // number_of_chunks) + 1
        this_offset = 0

        chunks = []
        src_block = array.array("B", block)
        dst_block = array.array("B", block)
        while this_offset < total_len:
            offset = dst_offset = this_offset
            if offset - overlap_each_block > 0:
                dst_offset = offset - overlap_each_block
            if this_offset + chunk_sz < total_len:
                length = chunk_sz
            else:
                length = total_len - this_offset
            chunks.append((offset, dst_offset, length))
            dst_block[dst_offset : dst_offset + length] = src_block[
                offset : offset + length
            ]
            this_offset += chunk_sz
        dst_len = dst_offset + length
        dst_block = dst_block[:dst_len].tobytes()

        fh_src, fh_dst = self._open_src_dst(src_filename, dst_filename)

        result = self.chan.copychunk(fh_src, fh_dst, chunks)
        self.assertEqual(result[0][0].chunks_written, number_of_chunks)
        self.assertEqual(result[0][0].total_bytes_written, total_len)

        # read each file and verify the result
        src_buf = self.chan.read(fh_src, total_len, 0).tobytes()
        self.assertBufferEqual(src_buf, block)
        dst_buf = self.chan.read(fh_dst, dst_len, 0).tobytes()
        self.assertBufferEqual(dst_buf, dst_block)

        self.chan.close(fh_src)
        self.chan.close(fh_dst)

    def test_overlap_multiple_chunks(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 10
        overlap = 64
        self.generic_ssc_overlap_test_case(block, num_of_chunks, overlap)

    def test_overlap_16_chunks_1024_overlap(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 16
        overlap = 1024
        self.generic_ssc_overlap_test_case(block, num_of_chunks, overlap)

    def test_overlap_15_chunks_4096_overlap(self):
        block = _gen_test_buffer(65535)
        num_of_chunks = 15
        overlap = 4096
        self.generic_ssc_overlap_test_case(block, num_of_chunks, overlap)

    def generic_ssc_negative_test_case(
        self,
        src_access=None,
        dst_access=None,
        src_disp=None,
        dst_disp=None,
        src_options=None,
        dst_options=None,
        src_brl=None,
        dst_brl=None,
        exp_error=None,
    ):
        block = b"Hello"
        total_len = len(block)
        src_filename = "src_negative.txt"
        dst_filename = "dst_negative.txt"
        self._create_and_write(src_filename, block)

        fh_src, fh_dst = self._open_src_dst(
            src_filename,
            dst_filename,
            src_access=src_access,
            src_disp=src_disp,
            src_options=src_options,
            dst_access=dst_access,
            dst_disp=dst_disp,
            dst_options=dst_options,
        )
        close_handles = []
        if src_brl or dst_brl:
            chan2, tree2 = self.tree_connect()
            if src_brl:
                fh_src_other = chan2.create(
                    tree2, src_filename, access=access_rwd, share=share_all
                ).result()
                chan2.lock(
                    fh_src_other, [(0, 2, pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK)]
                ).result()
                close_handles.append((chan2, fh_src_other))
            if dst_brl:
                fh_dst_other = chan2.create(
                    tree2, dst_filename, access=access_rwd, share=share_all
                ).result()
                chan2.lock(
                    fh_dst_other, [(3, 2, pike.smb2.SMB2_LOCKFLAG_EXCLUSIVE_LOCK)]
                ).result()
                close_handles.append((chan2, fh_dst_other))
        if exp_error is None:
            exp_error = pike.ntstatus.STATUS_SUCCESS
        try:
            with self.assert_error(exp_error):
                result = self.chan.copychunk(fh_src, fh_dst, [(0, 0, 5)])
        finally:
            for chan, fh in close_handles:
                chan.close(fh)
            self.chan.close(fh_src)
            self.chan.close(fh_dst)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_neg_src_exc_brl(self):
        """
        Initiate copychunk when another handle has an exclusive BRL on the
        source file (win 8 / 2012)
        """
        self.generic_ssc_negative_test_case(
            src_brl=True, exp_error=pike.ntstatus.STATUS_FILE_LOCK_CONFLICT
        )

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_neg_dst_exc_brl(self):
        """
        Initiate copychunk when another handle has an exclusive BRL on the
        destination file (win 8 / 2012)
        """
        self.generic_ssc_negative_test_case(
            dst_brl=True, exp_error=pike.ntstatus.STATUS_FILE_LOCK_CONFLICT
        )

    def test_neg_src_no_read(self):
        """
        Try to copychunk with no read access on the source file
        """
        self.generic_ssc_negative_test_case(
            src_access=pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE,
            exp_error=pike.ntstatus.STATUS_ACCESS_DENIED,
        )

    def test_neg_dst_no_read(self):
        """
        Try to copychunk with no read access on the destination file
        """
        self.generic_ssc_negative_test_case(
            dst_access=pike.smb2.FILE_WRITE_DATA | pike.smb2.DELETE,
            exp_error=pike.ntstatus.STATUS_ACCESS_DENIED,
        )

    def test_neg_dst_no_write(self):
        """
        Try to copychunk with no write access on the destination file
        """
        self.generic_ssc_negative_test_case(
            dst_access=pike.smb2.FILE_READ_DATA | pike.smb2.DELETE,
            exp_error=pike.ntstatus.STATUS_ACCESS_DENIED,
        )

    def test_neg_dst_is_a_dir(self):
        """
        Try to copychunk with destination file being a directory
        """
        self.generic_ssc_negative_test_case(
            dst_options=pike.smb2.FILE_DIRECTORY_FILE | pike.smb2.FILE_DELETE_ON_CLOSE,
            dst_disp=pike.smb2.FILE_OPEN_IF,
            exp_error=pike.ntstatus.STATUS_INVALID_DEVICE_REQUEST,
        )

    def generic_multiple_resume_key_test_case(self, sess_num):
        """
        multiple session resume key test
        """
        src_filename = "src_copy_chunk_rk.txt"
        content = "resume key test"
        rk_list = []

        for i in range(sess_num):
            chan, tree = self.tree_connect()
            fh_1 = chan.create(
                tree, src_filename, access=access_rwd, share=share_all
            ).result()
            chan.write(fh_1, 0, content)
            resume_key_1_1 = chan.resume_key(fh_1)[0][0].resume_key
            resume_key_1_2 = chan.resume_key(fh_1)[0][0].resume_key
            fh_2 = chan.create(
                tree, src_filename, access=access_rwd, share=share_all
            ).result()
            resume_key_2_1 = chan.resume_key(fh_2)[0][0].resume_key
            resume_key_2_2 = chan.resume_key(fh_2)[0][0].resume_key
            self.other_chan_list.append(
                {"channel": chan, "tree": tree, "filehandles": [fh_1, fh_2]}
            )
            try:
                # for same file handler, resume_key should be same
                self.assertEqual(resume_key_1_1, resume_key_1_2)
                self.assertEqual(resume_key_2_1, resume_key_2_2)
                # for different file handler in same session
                self.assertNotEqual(resume_key_1_1, resume_key_2_1)
                # for different file hander and different session
                self.assertNotIn(resume_key_1_1, rk_list)
                self.assertNotIn(resume_key_2_1, rk_list)
            except AssertionError:
                self.error(
                    "resume key check failed on session index {0}".format(i),
                    exc_info=True,
                )
                raise
            else:
                rk_list += [resume_key_1_1, resume_key_2_1]

    def test_multiple_resume_key(self):
        """
        test multiple sessions requesting resume_key for same
        file, also test within one session, multiple requests
        for same file
        """
        self.generic_multiple_resume_key_test_case(sess_num=3)

    def generic_negative_resume_key_test_case(self, resume_key, exp_error=None):
        """
        negative test for invalid resume key handling case
        """
        src_filename = "src_copy_chunk_rk.txt"
        dst_filename = "dst_copy_chunk_rk.txt"

        chunks = [(0, 0, len(SIMPLE_CONTENT))]
        fh_src, fh_dst = self._open_src_dst(src_filename, dst_filename)

        if exp_error is None:
            exp_error = pike.ntstatus.STATUS_SUCCESS
        try:
            with self.assert_error(exp_error):
                result = self.chan.copychunk(fh_src, fh_dst, chunks, resume_key)
        finally:
            self.chan.close(fh_src)
            self.chan.close(fh_dst)

    def test_other_resume_key(self):
        """
        negative test for multiple sessions acquiring resume key
        for same file one session uses other session's resume key
        to perform the server side copy
        """
        self._prepare_source_file()
        chan_2, tree_2 = self.tree_connect()
        src_filename = "src_copy_chunk_rk.txt"
        fh_src_2 = chan_2.create(
            tree_2, src_filename, access=access_rwd, share=share_all
        ).result()
        self.other_chan_list.append(
            {"channel": chan_2, "tree": tree_2, "filehandles": [fh_src_2]}
        )
        other_key = chan_2.resume_key(fh_src_2)[0][0].resume_key

        self.generic_negative_resume_key_test_case(
            other_key, exp_error=self.bad_resume_key_error
        )

    def test_bogus_resume_key(self):
        """
        negative test for single session does not acquire resume_key
        but generate one bogus resume_key to perform the server side copy
        """
        self._prepare_source_file()
        import array

        bogus_key = array.array("B", [255] * 24)
        self.generic_negative_resume_key_test_case(
            bogus_key, exp_error=self.bad_resume_key_error
        )

    def basic_ssc_random_test_case(self, chunks, src_options=0):
        """
        given chunks, perform server side copy accordingly
        """
        src_filename = "src_copy_chunk_random.txt"
        dst_filename = "dst_copy_chunk_random.txt"

        fh_src, fh_dst = self._open_src_dst(
            src_filename, dst_filename, src_options=src_options
        )

        result = self.chan.copychunk(fh_src, fh_dst, chunks)
        copy_len = chunks[-1][0] + chunks[-1][2]
        dst_offset = chunks[0][1]
        self.assertEqual(result[0][0].chunks_written, len(chunks))
        self.assertEqual(result[0][0].total_bytes_written, copy_len)

        # read each file and verify the result
        src_buf = self._read_big_file(fh_src, copy_len)
        dst_buf = self._read_big_file(fh_dst, copy_len, add_offset=dst_offset)
        self.assertBufferEqual(dst_buf, src_buf)

        self.chan.close(fh_src)
        self.chan.close(fh_dst)

    def generic_ssc_random_test_case_loop(self, **kwargs):
        """
        test loop for certain random parameters
        """
        src_filename = "src_copy_chunk_random.txt"
        self._gen_16mega_file(src_filename)
        iter_num = 3
        for i in range(iter_num):
            self.logger.info("iter %d, parameters flags are %s" % (i, kwargs))
            chunks = self._gen_random_chunks_on_para(**kwargs)
            if i == (iter_num - 1):
                src_options = pike.smb2.FILE_DELETE_ON_CLOSE
            else:
                src_options = 0
            self.basic_ssc_random_test_case(chunks, src_options=src_options)

    def test_random_copy_length(self):
        self.generic_ssc_random_test_case_loop(random_chunk_size=False)

    def test_random_chunk_size(self):
        self.generic_ssc_random_test_case_loop(random_chunk_size=True)

    def test_random_file_offset(self):
        self.generic_ssc_random_test_case_loop(
            random_chunk_size=False, random_offset=True
        )

    def test_random_all(self):
        self.generic_ssc_random_test_case_loop(
            random_chunk_size=True, random_offset=True
        )

    def generic_ssc_compound_test_case(self):
        """
        test copychunk request inside one compound smb message
        1st chunkcopy from src to dst file
        2nd read half of the dst file
        3rd read the other half of the dst file
        4st close the open
        """
        src_filename = "src_ssc_chunk_cp.txt"
        dst_filename = "dst_ssc_chunk_cp.txt"
        block = _gen_test_buffer(SIMPLE_5_CHUNKS_LEN)
        self._create_and_write(src_filename, block)
        fh_src, fh_dst = self._open_src_dst(src_filename, dst_filename)
        ioctl_req = self.chan.copychunk_request(fh_src, fh_dst, SIMPLE_5_CHUNKS)
        nb_req = ioctl_req.parent.parent
        read_req1 = self.chan.read_request(
            pike.model.RelatedOpen(self.tree), SIMPLE_5_CHUNKS_LEN // 2, 0
        )
        nb_req.adopt(read_req1.parent)
        read_req2 = self.chan.read_request(
            pike.model.RelatedOpen(self.tree),
            SIMPLE_5_CHUNKS_LEN // 2,
            SIMPLE_5_CHUNKS_LEN // 2,
        )
        nb_req.adopt(read_req2.parent)
        close_req = self.chan.close_request(pike.model.RelatedOpen(self.tree))
        nb_req.adopt(close_req.parent)

        (ssc_res, read1_res, read2_res, close_res) = self.chan.connection.transceive(
            nb_req
        )

        self.assertEqual(ssc_res[0][0].chunks_written, len(SIMPLE_5_CHUNKS))
        self.assertEqual(ssc_res[0][0].total_bytes_written, SIMPLE_5_CHUNKS_LEN)
        dst_buf = read1_res[0].data.tobytes() + read2_res[0].data.tobytes()
        self.assertBufferEqual(dst_buf, block)
        self.chan.close(fh_src)

    def test_ssc_in_compound_req(self):
        self.generic_ssc_compound_test_case()

    def generic_mchan_ssc_test_case(self):
        """
        Interaction with multiple channel feature:
        reqeust resume key in one channel
        copy chunk in another channel with the resume key
        """
        src_filename = "src_copy_chunk_mc.txt"
        dst_filename = "dst_copy_chunk_mc.txt"
        block = _gen_test_buffer(SIMPLE_5_CHUNKS_LEN)
        self._create_and_write(src_filename, block)
        chan2 = (
            self.chan.connection.client.connect(self.server)
            .negotiate()
            .session_setup(self.creds, bind=self.chan.session)
        )

        fh_src, fh_dst = self._open_src_dst(src_filename, dst_filename)
        resume_key = self.chan.resume_key(fh_src)[0][0].resume_key
        result = chan2.copychunk(fh_src, fh_dst, SIMPLE_5_CHUNKS, resume_key)

        self.assertEqual(result[0][0].chunks_written, len(SIMPLE_5_CHUNKS))
        self.assertEqual(result[0][0].total_bytes_written, SIMPLE_5_CHUNKS_LEN)

        # read each file and verify the result
        src_buf = self.chan.read(fh_src, SIMPLE_5_CHUNKS_LEN, 0).tobytes()
        self.assertBufferEqual(src_buf, block)
        dst_buf = chan2.read(fh_dst, SIMPLE_5_CHUNKS_LEN, 0).tobytes()
        self.assertBufferEqual(dst_buf, block)
        self.chan.close(fh_src)
        self.chan.close(fh_dst)
        chan2.connection.close()

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    @pike.test.RequireCapabilities(pike.smb2.SMB2_GLOBAL_CAP_MULTI_CHANNEL)
    def test_ssc_in_multchannel(self):
        self.generic_mchan_ssc_test_case()

    def generic_multiple_ssc_test_case(
        self, filepair, chunks, blocks, write_thru=False, num_iter=1
    ):
        """
        server side copy in multiple sessions, multiple iterations
        with filepair, chunks, blocks, write_through flag as input
        parameters
        """
        num_sess = len(filepair)
        results = [None] * num_sess * num_iter
        ssc_futures = [None] * num_sess * num_iter

        self._prepare_before_ssc_copy(filepair, write_thru=write_thru)

        # for loop to submit requests in short time
        for it in range(num_iter):
            for i in range(num_sess):
                chunk_break = list(zip(*chunks))
                dst_list = [x + filepair[i][2] for x in chunk_break[1]]
                newchunks = list(zip(chunk_break[0], dst_list, chunk_break[2]))
                ssc_futures[
                    it * num_sess + i
                ] = self._submit_ssc_async_request_in_other_channels(i, newchunks)

        # collect asychronized results
        for i in range(num_sess * num_iter):
            results[i] = ssc_futures[i][0].result()
            copy_len = chunks[-1][0] + chunks[-1][2]
            self.assertEqual(results[i][0][0].chunks_written, len(chunks))
            self.assertEqual(results[i][0][0].total_bytes_written, copy_len)
            if i < num_sess:
                dst_buf = (
                    self.other_chan_list[i]["channel"]
                    .read(
                        self.other_chan_list[i]["filehandles"][1],
                        copy_len,
                        filepair[i][2],
                    )
                    .tobytes()
                )
                self.assertBufferEqual(dst_buf, blocks[i])

    def test_multiple_ssc_file(self):
        """
        server side copy with certain iterations in certain number
        of sessions
        """
        num_sess = 5
        num_iter = 3
        # file pair element {src_filename, dst_filename, dst_offset}
        filepair = []
        blocks = []
        for i in range(0, num_sess):
            src_filename = "src_multiple_session" + str(i) + ".txt"
            dst_filename = "dst_multiple_session" + str(i) + ".txt"
            filepair.append((src_filename, dst_filename, 0))
            block = _gen_random_test_buffer(SIMPLE_5_CHUNKS_LEN)
            self._create_and_write(src_filename, block)
            blocks.append(block)
        self.logger.info(
            "start multiple session test, session number is %d iteration number is %d"
            % (num_sess, num_iter)
        )
        self.generic_multiple_ssc_test_case(
            filepair, SIMPLE_5_CHUNKS, blocks, num_iter=num_iter
        )

    def test_multiple_ssc_same_source_file(self):
        """
        multiple server side copy operation which shares the same
        source file
        """
        num_sess = 5
        num_iter = 3
        # file pair element {src_filename, dst_filename, dst_offset}
        filepair = []
        blocks = []
        for i in range(0, num_sess):
            src_filename = "src_multiple_session_same_source.txt"
            dst_filename = "dst_multiple_session" + str(i) + ".txt"
            filepair.append((src_filename, dst_filename, 0))
            if i == 0:
                block = _gen_random_test_buffer(SIMPLE_5_CHUNKS_LEN)
                self._create_and_write(src_filename, block)
        blocks = [block] * num_sess
        self.logger.info(
            "ssc same source multiple session test, session number is %d iteration number is %d"
            % (num_sess, num_iter)
        )
        self.generic_multiple_ssc_test_case(
            filepair, SIMPLE_5_CHUNKS, blocks, num_iter=num_iter
        )

    def test_multiple_ssc_same_dest_file(self):
        """
        multiple server side copy operation which shares the same
        destination file but with different destination offset
        """
        num_sess = 5
        num_iter = 3
        # file pair element {src_filename, dst_filename, dst_offset}
        filepair = []
        blocks = []
        for i in range(0, num_sess):
            src_filename = "src_multiple_session" + str(i) + ".txt"
            dst_filename = "dst_multiple_session_same_dest.txt"
            filepair.append((src_filename, dst_filename, 0 + i * SIMPLE_5_CHUNKS_LEN))
            block = _gen_random_test_buffer(SIMPLE_5_CHUNKS_LEN)
            self._create_and_write(src_filename, block)
            blocks.append(block)
        self.logger.info(
            "ssc same dest multiple session test, session number is %d iteration number is %d"
            % (num_sess, num_iter)
        )
        self.generic_multiple_ssc_test_case(
            filepair, SIMPLE_5_CHUNKS, blocks, num_iter=num_iter
        )

    def test_multiple_ssc_file_with_writethrough_flag(self):
        """
        multiple server side copy operation which write_through
        flag set when creating the destination file handle
        """
        num_sess = 3
        num_iter = 1
        # file pair element {src_filename, dst_filename, dst_offset}
        filepair = []
        blocks = []
        for i in range(0, num_sess):
            src_filename = "src_multiple_session" + str(i) + ".txt"
            dst_filename = "dst_multiple_session" + str(i) + ".txt"
            filepair.append((src_filename, dst_filename, 0))
            block = _gen_random_test_buffer(SIMPLE_5_CHUNKS_LEN)
            self._create_and_write(src_filename, block)
            blocks.append(block)
        self.logger.info(
            "ssc test with write_through set, session number is %d iteration number is %d"
            % (num_sess, num_iter)
        )
        self.generic_multiple_ssc_test_case(
            filepair, SIMPLE_5_CHUNKS, blocks, write_thru=True, num_iter=num_iter
        )

    def generic_ssc_boundary_test_case(
        self, block, chunks, exp_error=None, ssc_res=[0, 0, 0]
    ):
        """
        check situation that one of the chunks which contain numbers cross the value
        of end of file or maximum values system can support
        """
        src_filename = "src_copy_chunk_cross_offset.txt"
        dst_filename = "dst_copy_chunk_cross_offset.txt"
        self._create_and_write(src_filename, block)

        fh_src, fh_dst = self._open_src_dst(src_filename, dst_filename)
        resume_key = self.chan.resume_key(fh_src)[0][0].resume_key
        future = self._submit_ssc_async_request(self.chan, fh_dst, resume_key, chunks)

        if exp_error is None:
            exp_error = pike.ntstatus.STATUS_SUCCESS
        try:
            with self.assert_error(exp_error):
                result = future[0].result()
        finally:
            self.chan.close(fh_src)
            self.chan.close(fh_dst)

        ioctl_resp = future[0].response.response.children[0]
        self.assertIsInstance(ioctl_resp.ioctl_output, pike.smb2.CopyChunkCopyResponse)
        chunks_written = ioctl_resp.children[0].chunks_written
        chunk_bytes_written = ioctl_resp.children[0].chunk_bytes_written
        total_bytes_written = ioctl_resp.children[0].total_bytes_written
        self.assertEqual(chunks_written, ssc_res[0])
        self.assertEqual(chunk_bytes_written, ssc_res[1])
        self.assertEqual(total_bytes_written, ssc_res[2])

    def test_neg_cross_offset_basic(self):
        block = _gen_test_buffer(4096)
        chunks = [(10, 0, 4096)]
        self.generic_ssc_boundary_test_case(
            block, chunks, exp_error=pike.ntstatus.STATUS_INVALID_VIEW_SIZE
        )

    def test_neg_cross_offset_second_chunk(self):
        block = _gen_test_buffer(8192)
        chunks = [(0, 0, 4096), (4097, 4096, 4096)]
        exp_res = [1, 0, 4096]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_VIEW_SIZE,
            ssc_res=exp_res,
        )

    def test_neg_cross_offset_fifth_chunk(self):
        block = _gen_test_buffer(20480)
        chunks = [
            (0, 0, 4096),
            (4096, 4096, 4096),
            (8192, 8192, 4096),
            (12288, 12288, 4096),
            (16386, 16384, 4096),
        ]
        exp_res = [4, 0, 16384]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_VIEW_SIZE,
            ssc_res=exp_res,
        )

    def test_neg_cross_max_chunks(self):
        """
        request contains 300 chunks
        """
        block = _gen_test_buffer(30000)
        chunks = [(0, 0, 100)] * 300
        exp_res = [self.max_number_of_chunks, self.max_chunk_size, self.max_data_size]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_PARAMETER,
            ssc_res=exp_res,
        )

    def test_neg_cross_chunk_size(self):
        """
        request with chunk size in first chunk larger
        than the maximum chunk size
        """
        block = _gen_test_buffer(1048577)
        chunks = [(0, 0, 1048577)]
        exp_res = [self.max_number_of_chunks, self.max_chunk_size, self.max_data_size]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_PARAMETER,
            ssc_res=exp_res,
        )

    def test_neg_cross_second_chunk_size(self):
        """
        request with second chunk chunk size larger
        than the maximum chunk size
        """
        block = _gen_test_buffer(2097153)
        chunks = [(0, 0, 1048576), (1048576, 1048576, 1048577)]
        exp_res = [self.max_number_of_chunks, self.max_chunk_size, self.max_data_size]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_PARAMETER,
            ssc_res=exp_res,
        )

    def test_neg_cross_chunk_size_allf(self):
        """
        request with chunk size in one chunk all f
        """
        block = _gen_test_buffer(1048577)
        chunks = [(0, 0, 4294967295)]
        exp_res = [self.max_number_of_chunks, self.max_chunk_size, self.max_data_size]
        self.generic_ssc_boundary_test_case(
            block,
            chunks,
            exp_error=pike.ntstatus.STATUS_INVALID_PARAMETER,
            ssc_res=exp_res,
        )



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        queryondiskid.py
#
# Abstract:
#
#        Query On Disk Id CreateContext tests
#
# Authors: Masen Furer <masen.furer@emc.com>
#

import array

import pike.model
import pike.smb2
import pike.test

share_all = (
    pike.smb2.FILE_SHARE_READ | pike.smb2.FILE_SHARE_WRITE | pike.smb2.FILE_SHARE_DELETE
)
access_rwd = pike.smb2.GENERIC_READ | pike.smb2.GENERIC_WRITE | pike.smb2.DELETE
null_fid = array.array("B", [0] * 32)


class TestQueryOnDiskID(pike.test.PikeTest):
    test_files = [
        "qfid_file.bin",
        "qfid_same_file.bin",
        "qfid_diff_file1.bin",
        "qfid_diff_file2.bin",
    ]

    def extract_file_id(self, response):
        """
        Pull out the Query On Disk ID file_id field and return it
        """
        create_res = qfid_res = None
        for child in response:
            if isinstance(child, pike.smb2.CreateResponse):
                create_res = child
                break
        self.assertIsNotNone(
            create_res,
            "response didn't contain a " "CreateResponse: {0}".format(response),
        )
        for child in create_res:
            if isinstance(child, pike.smb2.QueryOnDiskIDResponse):
                qfid_res = child
                break
        self.assertIsNotNone(
            qfid_res,
            "CreateResponse didn't contain a "
            "QueryOnDiskIDResponse: {0}".format(create_res),
        )
        return qfid_res.file_id

    def test_qfid_functional(self):
        """
        Sending a QFid request should ellicit a success response
        """

        chan, tree = self.tree_connect()
        open_future = chan.create(
            tree,
            "qfid_file.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh = open_future.result()
        fid = self.extract_file_id(open_future.request_future.result())
        self.assertNotEqual(fid, null_fid, "On disk file_id was null")

    def test_qfid_same_file(self):
        """
        Opening the same file from different sessions should return the same id
        """

        chan1, tree1 = self.tree_connect()
        chan2, tree2 = self.tree_connect()
        open_future1 = chan1.create(
            tree1,
            "qfid_same_file.bin",
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh1 = open_future1.result()
        fid1 = self.extract_file_id(open_future1.request_future.result())
        open_future2 = chan2.create(
            tree2,
            "qfid_same_file.bin",
            access=access_rwd,
            share=share_all,
            disposition=pike.smb2.FILE_OPEN,
            query_on_disk_id=True,
        )
        fh2 = open_future2.result()
        fid2 = self.extract_file_id(open_future2.request_future.result())
        self.assertNotEqual(fid1, null_fid, "On disk file_id was null")
        self.assertEqual(fid1, fid2, "On disk file_id for same file didn't match")

    def test_qfid_diff_file(self):
        """
        Opening a different file from different sessions should NOT return
        the same id
        """

        chan1, tree1 = self.tree_connect()
        chan2, tree2 = self.tree_connect()
        open_future1 = chan1.create(
            tree1,
            "qfid_diff_file1.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh1 = open_future1.result()
        fid1 = self.extract_file_id(open_future1.request_future.result())
        open_future2 = chan2.create(
            tree2,
            "qfid_diff_file2.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh2 = open_future2.result()
        fid2 = self.extract_file_id(open_future2.request_future.result())
        self.assertNotEqual(fid1, null_fid, "On disk file_id was null")
        self.assertNotEqual(
            fid1, fid2, "On disk file_id for different files was the same"
        )

    def test_qfid_same_file_seq(self):
        """
        Opening the same file name after closing the first file should return
        the same on disk id
        """

        chan, tree = self.tree_connect()
        open_future = chan.create(
            tree,
            "qfid_file.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_SUPERSEDE,
            query_on_disk_id=True,
        )
        fh = open_future.result()
        fid = self.extract_file_id(open_future.request_future.result())
        self.assertNotEqual(fid, null_fid, "On disk file_id was null")
        first_fid = fid
        chan.close(fh)

        open_future = chan.create(
            tree,
            "qfid_file.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_OPEN,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh = open_future.result()
        fid = self.extract_file_id(open_future.request_future.result())
        self.assertNotEqual(fid, null_fid, "On disk file_id was null")
        self.assertEqual(fid, first_fid, "Subsequent open returns different file id")

    def test_qfid_same_file_seq_delete(self):
        """
        Opening the same file name after deleting the first file should return
        a different on disk id
        """

        chan, tree = self.tree_connect()
        open_future = chan.create(
            tree,
            "qfid_file.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_SUPERSEDE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh = open_future.result()
        fid = self.extract_file_id(open_future.request_future.result())
        self.assertNotEqual(fid, null_fid, "On disk file_id was null")
        first_fid = fid
        chan.close(fh)

        open_future = chan.create(
            tree,
            "qfid_file.bin",
            access=access_rwd,
            disposition=pike.smb2.FILE_CREATE,
            options=pike.smb2.FILE_DELETE_ON_CLOSE,
            query_on_disk_id=True,
        )
        fh = open_future.result()
        fid = self.extract_file_id(open_future.request_future.result())
        self.assertNotEqual(fid, null_fid, "On disk file_id was null")
        self.assertNotEqual(
            fid, first_fid, "Subsequent open after delete returns same file id"
        )



#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        transport.py
#
# Abstract:
#
#        Async event loop, socket handling, and polling mechanisms
#
# Authors: Masen Furer (masen.furer@dell.com)
#

from builtins import object
from errno import (
    errorcode,
    EBADF,
    ECONNRESET,
    ENOTCONN,
    ESHUTDOWN,
    ECONNABORTED,
    EISCONN,
    EINPROGRESS,
    EALREADY,
    EWOULDBLOCK,
    EAGAIN,
)
import select
import socket
import time

_reraised_exceptions = (KeyboardInterrupt, SystemExit)


class Transport(object):
    """
    Transport is responsible for managing the underlying socket, registering
    for socket events, dispatching read, write, and errors to higher layers.
    It is analogous to asyncore.dispatcher and is a drop in replacement for
    most purposes.

    If the alternate_poller is specified on instantiation, then the connection
    will register for events on that poller as opposed to the global poller.
    """

    def __init__(self, alternate_poller=None):
        self.addr = None
        self.connected = False
        self.socket = None
        self._fileno = None
        if alternate_poller is not None:
            self.poller = alternate_poller
        else:
            self.poller = poller  # global poller

    def create_socket(self, family, type):
        """
        Creates the underlying non-blocking socket and associates it with this
        Transport's underlying poller
        """
        self.family_and_type = family, type
        sock = socket.socket(family, type)
        sock.setblocking(0)
        self.set_socket(sock)

    def set_socket(self, sock):
        """
        mirror the given Socket sock's file descriptor on the Transport and
        register this Transport with the underlying poller
        """
        self.socket = sock
        self._fileno = sock.fileno()
        self.poller.add_channel(self)

    def connect(self, address):
        """
        begin establishing a connection to the (host, port) address tuple.

        must call create_socket first. if the underlying socket is non-blocking
        then this command may return before the connection is established.

        higher level code should wait for the handle_connect event to signal
        that the endpoint is successfully connected
        """
        self.connected = False
        err = self.socket.connect_ex(address)
        if err in (EINPROGRESS, EALREADY, EWOULDBLOCK):
            return
        if err in (0, EISCONN):
            self.addr = address
            self.handle_connect_event()
        else:
            raise socket.error(err, errorcode[err])

    def close(self):
        """
        close the underlying socket connection and unregister this Transport
        from the underlying poller
        """
        self.socket.close()
        self.connected = False
        self.poller.del_channel(self)

    def send(self, data):
        """
        send data bytes over the connection. if the socket would block,
        schedule this Transport to be notified when the socket is available
        for writing. handle_write will be called in this case.

        returns the number of bytes sent or zero if the write would block
        """
        result = 0
        try:
            result = self.socket.send(data)
        except socket.error as err:
            if err.errno == EAGAIN:
                # reschedule the send when the socket is ready
                self.poller.defer_write(self)
            else:
                # raise non-retryable errors
                raise
        return result

    def recv(self, bufsize):
        """
        recv bufsize bytes over the connection. if the socket would block,
        then return an empty buffer. When the socket is available for reading
        handle_read will be called.

        returns a string representing the bytes received
        """
        result = ""
        try:
            result = self.socket.recv(bufsize)
            if result == "":
                raise EOFError("Remote host closed connection")
        except socket.error as err:
            # raise non-retryable errors
            if err.errno != EAGAIN:
                raise
        return result

    def handle_connect_event(self):
        """
        called internally when the socket becomes connected
        """
        self.connected = True
        self.handle_connect()

    def handle_connect(self):
        """
        callback fired when connection is established
        """
        pass

    def handle_read(self):
        """
        callback fired when the socket has data available
        """
        pass

    def handle_write(self):
        """
        callback fired when the socket is available for writing

        note: unlike asyncore, write notifications are not provided by default.
        this is a performance optimization because the socket is usually
        available for writing, and the application usually knows when it wants
        to write. There is no point in filling the event queues with
        write ready messages that will be ignored if the client has no data to
        send.

        Instead, applications are expected to implement handle_write, but to
        call it directly when data is to be sent. IF the socket would block,
        EALREADY will be handled by the Transport. The Transport requests a
        single write notification from the pollwer; when received, handle_write
        will be called once signalling that the socket may now be ready to retry

        If the application would prefer to be notified when the socket is ready
        to write, transport.poller.defer_write(transport) may be called to
        schedule a single handle_write callback.
        """
        pass

    def handle_close(self):
        """
        callback fired when the socket is closed
        """
        pass

    def handle_error(self):
        """
        callback fired if a non-recoverable exception is raised
        """
        pass


class BasePoller(object):
    """
    A poller is an underlying event monitoring system. This generic class
    can be built upon to implement efficient file descriptor polling methods
    which are available on various platforms.

    A minimal subclass must implement the poll() function which performs a
    single iteration of the event loop across all monitored Transports and
    calls process_readables and process_writables with the correct values.

    Subclasses should, in most cases call, into BasePoller methods in order
    to maintain proper accounting structures. The exception is when the poller
    handles accounting itself.
    """

    def __init__(self):
        """
        initialize the poller and register any kernel global structures
        necessary to monitor the file descriptors
        """
        self.connections = {}
        self.deferred_writers = set()

    def add_channel(self, transport):
        """
        begin monitoring the transport socket for read/connect events

        the underlying poller should not monitor Transports for writability
        except when:

            * the Transport's connection has not yet been established
            * the Transport has been passed as an argument to defer_write
        """
        self.connections[transport._fileno] = transport
        transport.poller = self

    def del_channel(self, transport):
        """
        stop monitoring the transport socket
        """
        del self.connections[transport._fileno]

    def defer_write(self, transport):
        """
        defers a write on the given transport. once the async poller determines
        that the transport can be written to, handle_write will be called
        """
        self.deferred_writers.add(transport._fileno)

    def loop(self, timeout=None, count=None):
        """
        enter the async event loop for the given timeout or number of iterations
        """
        start = time.time()
        complete_iterations = 0
        while True:
            if count is not None and complete_iterations >= count:
                break
            self.poll()
            if timeout is not None and time.time() > start + timeout:
                break
            complete_iterations += 1

    def poll(self):
        """
        Must be implemented by subclasses to execute a single iteration of the
        event loop. Based on the outcome of the events, the following actions
        MUST be performed

            * process_readables is called with a list of file descriptors which
              have data available for reading
            * process_writables is called with a list of file descriptors which
              have data available for writing
        """
        raise NotImplementedError("BasePoller does not have a polling mechanism")

    def process_readables(self, readables):
        """
        call handle_read on each applicable fd in the readables sequence and
        subsequently handle_error if any exception is raised or handle_close
        if the underlying socket is no longer connected
        """
        for fileno in readables:
            t = self.connections[fileno]
            try:
                t.handle_read()
            except socket.error as e:
                if e.args[0] not in (
                    EBADF,
                    ECONNRESET,
                    ENOTCONN,
                    ESHUTDOWN,
                    ECONNABORTED,
                ):
                    t.handle_error()
                else:
                    t.handle_close()
            except _reraised_exceptions:
                raise
            except:
                t.handle_error()

    def process_writables(self, writables):
        """
        for each Transport t corresponding to an fd in the writables sequence,
            if t is not marked as connected, call handle_connect_event
            otherwise call handle_write and remove the Transport from the set
            of deferred writers
            process close and error events if exception is encountered
        """
        for fileno in writables:
            t = self.connections[fileno]
            try:
                if not t.connected:
                    t.handle_connect_event()
                else:
                    if fileno in self.deferred_writers:
                        self.deferred_writers.remove(fileno)
                    t.handle_write()
            except socket.error as e:
                if e.args[0] not in (
                    EBADF,
                    ECONNRESET,
                    ENOTCONN,
                    ESHUTDOWN,
                    ECONNABORTED,
                ):
                    t.handle_error()
                else:
                    t.handle_close()
            except _reraised_exceptions:
                raise
            except:
                t.handle_error()


class KQueuePoller(BasePoller):
    """
    Implementation of KQueue, available on Mac OS and BSD derivatives
    """

    def __init__(self):
        super(KQueuePoller, self).__init__()
        self.kq = select.kqueue()
        self.batch_size = 10

    def add_channel(self, transport):
        super(KQueuePoller, self).add_channel(transport)
        events = [
            select.kevent(
                transport._fileno,
                filter=select.KQ_FILTER_READ,
                flags=select.KQ_EV_ADD | select.KQ_EV_ENABLE,
            ),
            select.kevent(
                transport._fileno,
                filter=select.KQ_FILTER_WRITE,
                flags=(select.KQ_EV_ADD | select.KQ_EV_ENABLE | select.KQ_EV_ONESHOT),
            ),
        ]
        self.kq.control(events, 0)

    def defer_write(self, transport):
        super(KQueuePoller, self).defer_write(transport)
        events = [
            select.kevent(
                transport._fileno,
                filter=select.KQ_FILTER_WRITE,
                flags=(select.KQ_EV_ADD | select.KQ_EV_ENABLE | select.KQ_EV_ONESHOT),
            )
        ]
        self.kq.control(events, 0)

    def poll(self):
        events = self.kq.control(None, self.batch_size, 0)
        readables = []
        writables = []
        for ev in events:
            if ev.filter == select.KQ_FILTER_READ:
                readables.append(ev.ident)
            elif ev.filter == select.KQ_FILTER_WRITE:
                writables.append(ev.ident)
        self.process_readables(readables)
        self.process_writables(writables)


class SelectPoller(BasePoller):
    """
    Implementation of select, available on most platforms as a fallback.

    Roughly equivalent performance to using asyncore
    """

    def poll(self):
        non_connected = [
            t._fileno for t in self.connections.values() if not t.connected
        ]
        readers = list(self.connections.keys())
        writers = non_connected + list(self.deferred_writers)
        readables, writables, _ = select.select(readers, writers, [], 0)
        self.process_readables(readables)
        self.process_writables(writables)


class PollPoller(BasePoller):
    """
    Implementation of poll, available on Linux
    """

    def __init__(self):
        super(PollPoller, self).__init__()
        self.p = select.poll()
        self.read_events = (
            select.POLLIN
            | select.POLLERR
            | select.POLLHUP
            | select.POLLNVAL
            | select.POLLMSG
            | select.POLLPRI
        )
        self.write_events = select.POLLOUT

    def add_channel(self, transport):
        super(PollPoller, self).add_channel(transport)
        self.p.register(transport._fileno, self.read_events | self.write_events)

    def del_channel(self, transport):
        super(PollPoller, self).del_channel(transport)
        self.p.unregister(transport._fileno)

    def defer_write(self, transport):
        super(PollPoller, self).defer_write(transport)
        self.p.modify(transport._fileno, self.read_events | self.write_events)

    def poll(self):
        events = self.p.poll(0)
        readables = []
        writables = []
        for fd, event in events:
            if event & self.read_events:
                readables.append(fd)
            elif event & self.write_events:
                writables.append(fd)
                self.p.modify(fd, self.read_events)
        self.process_readables(readables)
        self.process_writables(writables)


# Global poller / loop function for simple use cases
# more advanced tests or frameworks may use a custom
# poller implementation by setting a poller object onto
# a group of transports.
# The global poller will use the best polling mechanism available on the system
if hasattr(select, "kqueue"):
    poller = KQueuePoller()
elif hasattr(select, "poll"):
    poller = PollPoller()
else:
    poller = SelectPoller()


def loop(timeout=None, count=None):
    poller.loop(timeout, count)



#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#

from __future__ import unicode_literals
from builtins import str

import array

import pytest

import pike.model
import pike.smb2


@pytest.fixture
def mock_chan():
    mock_chan = pike.model.Channel(None, None, None)
    mock_chan.request = lambda *args, **kwargs: pike.smb2.Smb2(None)
    return mock_chan


class MockOpen(object):
    file_id = None


@pytest.mark.parametrize(
    "buf,expected_exception",
    (
        (None, None),
        ("unicode", UnicodeWarning),
        (b"literal bytes", None),
        (array.array("B", b"byte array"), None),
        (array.array("b", b"little byte array"), ValueError),
        ([], TypeError),
        (["foo"], TypeError),
        ([b"foo"], TypeError),
        (False, TypeError),
        (True, TypeError),
        (42, TypeError),
    ),
    ids=(
        "none",
        "unicode",
        "bytes",
        "byte array",
        "signed byte array",
        "empty list",
        "list of unicode",
        "list of bytes",
        "false",
        "true",
        "int",
    ),
)
def test_write_request_buffer_type(mock_chan, buf, expected_exception, recwarn):
    try:
        write_req = mock_chan.write_request(MockOpen, 0, buf)
        if expected_exception is not None and expected_exception not in [
            wm.category for wm in recwarn.list
        ]:
            pytest.fail("Did not raise {!r}".format(expected_exception))
        if isinstance(buf, str):
            buf = buf.encode("ascii")
        assert write_req.buffer == buf
    except expected_exception:
        pass



#!/usr/bin/env python
#
# Copyright (c) 2016-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
import ctypes
import os

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError

_HERED = os.path.abspath(os.path.dirname(__file__))

try:
    libgssapi_krb5 = ctypes.CDLL("libgssapi_krb5.so")
    defines = [
        ("HAVE_GSS_SET_CRED_OPTION", hasattr(libgssapi_krb5, "gss_set_cred_option")),
        (
            "HAVE_GSSSPI_SET_CRED_OPTION",
            hasattr(libgssapi_krb5, "gssspi_set_cred_option"),
        ),
    ]
    lw_krb_module = Extension(
        "pike.kerberos",
        [
            "pykerb/base64.c",
            "pykerb/kerberosbasic.c",
            "pykerb/kerberos.c",
            "pykerb/kerberosgss.c",
            "pykerb/kerberospw.c",
        ],
        include_dirs=[os.path.join(_HERED, "pykerb", "vendor")],
        libraries=["gssapi_krb5"],
        define_macros=defines,
    )
    setup(ext_modules=[lw_krb_module])
except (OSError, CCompilerError, DistutilsExecError, DistutilsPlatformError):
    print("libgssapi_krb5 not available, skipping kerberos module")
    setup()



#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#

import pike.core as core
import pike.model as model
import pike.smb2 as smb2
import pike.test


class TestClientCallbacks(pike.test.PikeTest):
    def test_pre_serialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateRequest))
                self.assertFalse(hasattr(frame, "buf"))
                callback_future.complete(True)

        self.default_client.register_callback(model.EV_REQ_PRE_SERIALIZE, cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_post_serialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateRequest))
                self.assertTrue(hasattr(frame, "buf"))
                self.assertEqual(frame.len + 4, len(frame.buf))
                callback_future.complete(True)

        self.default_client.register_callback(model.EV_REQ_POST_SERIALIZE, cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_pre_post_send(self):
        pre_callback_future = model.Future()
        post_callback_future = model.Future()
        expected_bytes = []

        def pre_cb(data):
            with pre_callback_future:
                expected_bytes.insert(0, len(data))
                self.assertGreater(expected_bytes[0], 16)
                pre_callback_future.complete(True)

        def post_cb(bytes_written):
            with post_callback_future:
                self.assertEqual(expected_bytes.pop(), bytes_written)
                if not expected_bytes:
                    post_callback_future.complete(True)

        self.default_client.register_callback(model.EV_REQ_PRE_SEND, pre_cb)
        self.default_client.register_callback(model.EV_REQ_POST_SEND, post_cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(pre_callback_future.result(timeout=2))
        self.assertTrue(post_callback_future.result(timeout=2))

    def test_pre_deserialize(self):
        callback_future = model.Future()

        def cb(data):
            with callback_future:
                self.assertGreater(len(data), 4)
                callback_future.complete(True)

        self.default_client.register_callback(model.EV_RES_PRE_DESERIALIZE, cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_post_deserialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateResponse))
                self.assertTrue(hasattr(frame, "buf"))
                self.assertEqual(frame.len + 4, len(frame.buf))
                callback_future.complete(True)

        self.default_client.register_callback(model.EV_RES_POST_DESERIALIZE, cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_pre_post_recv(self):
        pre_callback_future = model.Future()
        post_callback_future = model.Future()
        expected_bytes = []
        expected_rounds = [1, 2]

        def pre_cb(read_bytes):
            with pre_callback_future:
                expected_bytes.insert(0, read_bytes)
                self.assertGreater(read_bytes, 3)
                pre_callback_future.complete(True)

        def post_cb(data):
            with post_callback_future:
                self.assertEqual(expected_bytes.pop(), len(data))
                expected_rounds.pop()
                if not expected_rounds:
                    post_callback_future.complete(True)

        self.default_client.register_callback(model.EV_RES_PRE_RECV, pre_cb)
        self.default_client.register_callback(model.EV_RES_POST_RECV, post_cb)
        conn = self.default_client.connect(self.server, self.port)
        conn.negotiate()
        self.assertTrue(pre_callback_future.result(timeout=2))
        self.assertTrue(post_callback_future.result(timeout=2))


class TestConnectionCallbacks(pike.test.PikeTest):
    def test_pre_serialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateRequest))
                self.assertFalse(hasattr(frame, "buf"))
                callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_REQ_PRE_SERIALIZE, cb)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_post_serialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateRequest))
                self.assertTrue(hasattr(frame, "buf"))
                self.assertEqual(frame.len + 4, len(frame.buf))
                callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_REQ_POST_SERIALIZE, cb)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_pre_post_send(self):
        pre_callback_future = model.Future()
        post_callback_future = model.Future()
        expected_bytes = []

        def pre_cb(data):
            with pre_callback_future:
                expected_bytes.insert(0, len(data))
                self.assertGreater(expected_bytes[0], 16)
                pre_callback_future.complete(True)

        def post_cb(bytes_written):
            with post_callback_future:
                self.assertEqual(expected_bytes.pop(), bytes_written)
                if not expected_bytes:
                    post_callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_REQ_PRE_SEND, pre_cb)
        conn.register_callback(model.EV_REQ_POST_SEND, post_cb)
        conn.negotiate()
        self.assertTrue(pre_callback_future.result(timeout=2))
        self.assertTrue(post_callback_future.result(timeout=2))

    def test_pre_deserialize(self):
        callback_future = model.Future()

        def cb(data):
            with callback_future:
                self.assertGreater(len(data), 4)
                callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_RES_PRE_DESERIALIZE, cb)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_post_deserialize(self):
        callback_future = model.Future()

        def cb(frame):
            with callback_future:
                self.assertTrue(isinstance(frame, core.Frame))
                self.assertTrue(isinstance(frame[0], smb2.Smb2))
                self.assertTrue(isinstance(frame[0][0], smb2.NegotiateResponse))
                self.assertTrue(hasattr(frame, "buf"))
                self.assertEqual(frame.len + 4, len(frame.buf))
                callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_RES_POST_DESERIALIZE, cb)
        conn.negotiate()
        self.assertTrue(callback_future.result(timeout=2))

    def test_pre_post_recv(self):
        pre_callback_future = model.Future()
        post_callback_future = model.Future()
        expected_bytes = []
        expected_rounds = [1, 2]

        def pre_cb(read_bytes):
            with pre_callback_future:
                expected_bytes.insert(0, read_bytes)
                self.assertGreater(read_bytes, 3)
                pre_callback_future.complete(True)

        def post_cb(data):
            with post_callback_future:
                self.assertEqual(expected_bytes.pop(), len(data))
                expected_rounds.pop()
                if not expected_rounds:
                    post_callback_future.complete(True)

        conn = self.default_client.connect(self.server, self.port)
        conn.register_callback(model.EV_RES_PRE_RECV, pre_cb)
        conn.register_callback(model.EV_RES_POST_RECV, post_cb)
        conn.negotiate()
        self.assertTrue(pre_callback_future.result(timeout=2))
        self.assertTrue(post_callback_future.result(timeout=2))
